/*****************************************************************************
*   $Id: dgexec.c,v 1.1.1.1 2000/06/22 15:38:39 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    techno@spirit.net.au
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: dgexec.c,v $
* Revision 1.1.1.1  2000/06/22 15:38:39  tekno
* Opened source
*
*
*****************************************************************************/
/* dgexec.c - execute parsed template program */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <stdarg.h>
#include <sys/types.h>
#include <regex.h>
#include "dglex.h"
#include "dgvar.h"
#include "dgdb.h"
#include "y.tab.h"
#include "dgfilt.h"

FILE * outfile;

expr_result error(tree t, char * msg, ...)
{
  va_list ap;
  char buf[256];
  static int nerr;
#define MAX_ERRORS 100
  
  if (nerr < MAX_ERRORS)
  {
    va_start(ap, msg);
    vsnprintf(buf, sizeof(buf), msg, ap);
    fprintf(stderr, "%s:%d: %s\n", t->filename, t->lineno, buf);
  }
  nerr++;
  if (nerr == MAX_ERRORS)
    fprintf(stderr, 
    "%d errors so far; further semantic errors will not be printed\n", nerr);
  return E_ERROR;
}

expr_result error_in_pattern(tree t, regex_t * preg, int code)
{
  char buf[1024];
  regerror(code, preg, buf, sizeof(buf));
  fprintf(stderr, "%s:%d: %s\n", t->filename, t->lineno, buf);
  return E_ERROR;
}

static char * cat_strings(char * s1, char * s2)
{
  int l1, l2;
  char * r = tsmalloc((l1 = strlen(s1))+(l2 = strlen(s2))+1);
  memcpy(r, s1, l1);
  memcpy(r + l1, s2, l2);
  r[l1+l2] = 0;
  return r;
}

static char * negate_string(char * s)
{
  char * r = tsmalloc(strlen(s)+1);
  char * t = r;
  while (*s)
    *(t++) = (char)(256 - (unsigned char)*(s++));
  *t = 0;
  return r;
}

static void promote(expr_result * er1, expr_result * er2,
                    int * ei1, int * ei2,
                    char ** es1, char ** es2)
{
  /* Promote string/int combination to two strings. */
  char buf[20];
  if (*er1 == *er2)
    return;
  if (*er1 == E_INT)
  {
    sprintf(buf, "%d", *ei1);
    *es1 = tsmalloc(strlen(buf)+1);
    strcpy(*es1, buf);
    *er1 = E_STRING;
  }
  else
  {
    sprintf(buf, "%d", *ei2);
    *es2 = tsmalloc(strlen(buf)+1);
    strcpy(*es2, buf);
    *er2 = E_STRING;
  }
}

static void assign_var(tree t, expr_result er, int ei, char * es)
{
  var * v = lookup_var(&t->u.var);
  set_var(v, er, ei, es);
}

static void assign_array(tree t, expr_result erx, 
             int eix, char * esx, expr_result er, int ei, char * es)
{
  var * v;
  
  if (t->u.var.var && !(t->u.var.var->flags & (VAR_ARRAY|VAR_SLICE)))
    error(t, "variable %s is defined scalar", t->u.var.var->name);
  else
  {
    v = lookup_var(&t->u.var);
    set_array(v, t, erx, eix, esx, er, ei, es);
  }  
}

expr_result eval_expr(int * ei, char ** es, tree t)
{
  extern expr_result invoke_method(int *, char **, tree);
  expr_result erl, err, erx;
  int eil, eir, eix;
  char * esl, * esr, * esx;
  var * v;
  char strict;
  regex_t preg;
  regmatch_t pm;
  int code;
  
  *ei = 1;
  *es = "1";
  if (!t)
    return E_INT;
  switch (t->type)
  {
  case T_INT:
    *ei = t->u.intval;
    return E_INT;
  case T_STRING:
    *es = t->u.strval;
    return E_STRING;
  case T_VAR:
    v = lookup_var(&t->u.var);
    *ei = v->intval;
    *es = v->strval;
    return v->type;
  case T_METHOD:
    return invoke_method(ei, es, t);
    break;
  case T_UNARY:
    erl = eval_expr(&eil, &esl, t->u.expr.right);
    switch (t->u.expr.oper)
    {
    case '-':
      if (erl == E_INT)
      { *ei = -eil; return E_INT; }
      if (erl == E_STRING)
      { *es = negate_string(esl); return E_STRING; }
      break;
    case '!':
      if (erl == E_INT)
      { *ei = !eil; return E_INT; }
      if (erl == E_STRING)
      { *ei = !*esl; return E_INT; }
      break;
    case '~':
      if (erl == E_INT)
      { *ei = ~eil; return E_INT; }
      if (erl == E_STRING)
        return error(t, "Unary ~ used on string operand");
      break;
    }
    break;
  case T_INDEX:
    erx = eval_expr(&eix, &esx, t->u.expr.right);
    v = lookup_array_var(&t->u.expr.left->u.var);
    if (!(v->flags & (VAR_ARRAY|VAR_SLICE)))
      return error(t, "array index on uninit/scalar variable %s", v->name);
    return array_lookup(ei, es, v, t, erx, eix, esx);
    break;
  case T_BINARY:
    erl = eval_expr(&eil, &esl, t->u.expr.left);
    if (t->u.expr.oper != LOGAND && t->u.expr.oper != LOGOR)
    {
      /* For non-boolean ops, always evaluate RHS of operator  */
      err = eval_expr(&eir, &esr, t->u.expr.right);
      promote(&erl, &err, &eil, &eir, &esl, &esr);
    }
    else if (erl == E_STRING)
    {
      /* For boolean ops, coerce LHS string to int operand. */
      erl = E_INT;
      eil = esl[0];
    }
    if (erl == E_INT) switch (t->u.expr.oper)
    {
    case '*':
    case '#':	/* Concat equiv to multiplication for ints */
      *ei = eil * eir;
      break;
    case '/':
      if (!eir)
        return error(t, "div by 0");
      *ei = eil / eir;
      break;
    case '%':
      if (!eir)
        return error(t, "mod by 0");
      *ei = eil % eir;
      break;
    case '+':
      *ei = eil + eir;
      break;
    case '-':
      *ei = eil - eir;
      break;
    case SHL:
      *ei = eil << eir;
      break;
    case SHR:
      *ei = eil >> eir;
      break;
    case '<':
      *ei = eil < eir;
      break;
    case '>':
      *ei = eil > eir;
      break;
    case GE:
      *ei = eil >= eir;
      break;
    case LE:
      *ei = eil <= eir;
      break;
    case EQ:
      *ei = eil == eir;
      break;
    case NE:
      *ei = eil != eir;
      break;
    case '&':
      *ei = eil & eir;
      break;
    case '^':
      *ei = eil ^ eir;
      break;
    case '|':
      *ei = eil | eir;
      break;
    case LOGAND:
      *ei = !!eil;
      if (!eil)
        break;
      err = eval_expr(&eir, &esr, t->u.expr.right);
      if (err == E_ERROR) break;
      if (err == E_INT)
        *ei = !!eir;
      else
        *ei = !!esr[0];
      break;
    case LOGOR:
      *ei = !!eil;
      if (eil)
        break;
      err = eval_expr(&eir, &esr, t->u.expr.right);
      if (err == E_ERROR) break;
      if (err == E_INT)
        *ei = !!eir;
      else
        *ei = !!esr[0];
      break;
    default:
      return error(t, "bad operands");
    }
    else if (erl == E_STRING) switch (t->u.expr.oper)
    {
    case '+':
    case '#':	/* String concatentation */
      *es = cat_strings(esl, esr);
      return E_STRING;
      break;
    case '<':
      *ei = strcmp(esl, esr) < 0;
      break;
    case '>':
      *ei = strcmp(esl, esr) > 0;
      break;
    case GE:
      *ei = strcmp(esl, esr) >= 0;
      break;
    case LE:
      *ei = strcmp(esl, esr) <= 0;
      break;
    case EQ:
      *ei = strcmp(esl, esr) == 0;
      break;
    case NE:
      *ei = strcmp(esl, esr) != 0;
      break;
    case MATCHES:
      strict = 0;
      goto do_rx;
      break;
    case STRICTMATCHES:
      strict = 1;
    do_rx:
      if ((code = regcomp(&preg, esr, REG_EXTENDED | REG_NEWLINE)))
        return error_in_pattern(t, &preg, code);
      *ei = regexec(&preg, esl, 1, &pm, 0);
      regfree(&preg);
      if (*ei)
      {
        /* No match or error */
        if (*ei == REG_NOMATCH)
        {
          *ei = 0;
          break;
        }
        else
          return error(t, "internal error in regexec");
      }
      /* Match, determine if strict */
      *ei = 1;
      if (strict && (pm.rm_so || pm.rm_eo != strlen(esl)))
        *ei = 0;
      break;
    default:
      return error(t, "bad operands");
    }
    return E_INT;
  default:
    break;
  }
  fprintf(stderr, "eval_expr bad type: t->type = %d\n", t->type);
  return error(t, "internal");
}

/* Statics used to interface into qsort callback */
static int nsort;		/* Number of sort expressions */
static tree sortexprs[32];	/* Pointers to each */
static expr_result sorttypes[32];  /* Type of result of sort expr */
static var * loopvar;		/* Current forall var */
static int sortobjsize;		/* Size of sort objects created */

static object * setup_sort(list sl, var * v, long nobj)
{
  nsort = 0;
  while (sl)
  {
    sorttypes[nsort] = E_VAR;	/* Set to this until actually known */
    sortexprs[nsort++] = sl->val;
    sl = sl->next;
  }
  loopvar = v;
  /* A sort object is an object, with the 'next' pointer used to point
     to the actual object data, and resolved used to indicate whether
     the corresponding expression has been evaluated.  data entries
     contain the result of the expression. */
  sortobjsize = offsetof(object, data) + sizeof(char *)*nsort;
  if (nobj)
  {
    object * sot = (object *)calloc(nobj, sortobjsize);
    return sot;
  }
  return 0; 
}

static int sort_comp(object * a, object * b)
{
  /* qsort callback.  Compare two sortobjs.  Actually, we sort in descending
     order because the resultant array is scanned backwards. */
  int n;
  expr_result er;
  int ei;
  char * es;
  int ret;
  
  for (n = 0; n < nsort; n++)
  {
    if (!(a->resolved & 1 << n))
    {
      set_forallvar(loopvar, a->next);
      er = eval_expr(&ei, &es, sortexprs[n]);
      if (er != E_ERROR && sorttypes[n] == E_VAR)
        sorttypes[n] = er;
      if (sorttypes[n] == E_STRING)
        a->data[n] = es;
      else
        a->data[n] = (char *)ei;
      a->resolved |= 1 << n;
    }
    if (!(b->resolved & 1 << n))
    {
      set_forallvar(loopvar, b->next);
      er = eval_expr(&ei, &es, sortexprs[n]);
      if (er != E_ERROR && sorttypes[n] == E_VAR)
        sorttypes[n] = er;
      if (sorttypes[n] == E_STRING)
        b->data[n] = es;
      else
        b->data[n] = (char *)ei;
      b->resolved |= 1 << n;
    }
    
    if (sorttypes[n] == E_STRING)
      ret = strcmp(b->data[n], a->data[n]);
    else
      ret = (int)b->data[n] - (int)a->data[n];
    if (ret)
      return ret;
  }
  return 0;
}

static void keep(object * sot, long * nobj, int ssize, tree subset, var * v,
                 int nsort, int (*sort_comp)(const void *, const void *))
{
  /* Condense the sot list according to subset constraints (if any). 
     Then sort the list. */
  object * s;
  long i, j;
  expr_result er;
  int ei;
  char * es;
  
  if (subset)
  {
    for (s = sot, i = 0, j = 0; 
         i < *nobj; 
         i++, s = (object *)((char *)s + ssize))
    {
      set_forallvar(v, s->next);
      er = eval_expr(&ei, &es, subset);
      if (er == E_INT && ei)
      {
        if (i > j)
          memcpy((char *)sot + ssize*j, (char *)sot + ssize*i, ssize);
        j++;
      }
    }
    *nobj = j;
  }
  if (nsort && *nobj)
    qsort(sot, *nobj, ssize, sort_comp);
}

void recycle_stmt(tree t);

void recycle_list(list l)
{
  list ln;
  
  while (l)
  {
    ln = l->next;
    recycle_stmt(l->val);
    free(l);
    l = ln;
  }
}

void recycle_stmt(tree t)
{
  /* Called to discard statement resources after immediate execution.
     Recurse through tree, unreference variables, free strings and list/tree
     nodes. */
  var * v;
  
  if (!t)
    return;
  
  switch (t->type)
  {
  case T_INT:
  case T_FLOAT:
    break;
  case T_STRING:
    free(t->u.strval);
    break;
  case T_SET:
  case T_EXPR:
  case T_PARM:
  case T_UNARY:
  case T_BINARY:
  case T_INDEX:
  case T_REFERS:
    recycle_stmt(t->u.expr.left);
    recycle_stmt(t->u.expr.right);
    break;
  case T_REGEXP:
    free(t->u.rx);
    break;
  case T_OUTFILE:
    /* Can't free outfile, since it is needed for subsequent output */
    return;
  case T_VAR:
    v = delete_var_ref(&t->u.var);
    /* Only free if not used by var itself */
    if (!v || v->name != t->u.var.varname)
      free(t->u.var.varname);
    break;
  case T_METHOD:
    recycle_stmt(t->u.meth.obj);
    recycle_list(t->u.meth.parms);
    break;
  case T_IF:
    recycle_list(t->u.cond.conds);
    break;
  case T_COND:
    recycle_stmt(t->u.cn.cond);
    recycle_list(t->u.cn.stmts);
    break;
  case T_SWITCH:
    recycle_stmt(t->u.sw.expr);
    recycle_list(t->u.sw.conds);
    break;
  case T_FOR:
    recycle_stmt(t->u.loop.test);
    recycle_list(t->u.loop.inits);
    recycle_list(t->u.loop.assigns);
    recycle_list(t->u.loop.stmts);
    break;
  case T_FORALL:
    recycle_stmt(t->u.forall.datatype);
    recycle_stmt(t->u.forall.var);
    recycle_stmt(t->u.forall.keep);
    recycle_list(t->u.forall.sort);
    recycle_list(t->u.forall.ifnone);
    recycle_list(t->u.forall.stmts);
    break;
  case T_FUNCTION:
    recycle_stmt(t->u.mdef.thistype);
    recycle_stmt(t->u.mdef.ret);
    recycle_list(t->u.mdef.parms);
    recycle_list(t->u.mdef.stmts);
    free(t->u.mdef.name);
    break;
  default:
    /* Shouldn't happen */
    fprintf(stderr, "%s:%d: recycle() unknown type %d\n",
      t->filename, t->lineno, t->type);
    return;
  }
  free(t);
}

int exec_stmt(tree t)
{
  expr_result er, erx;
  int ret = 0;	/* Return val from this fn */
  int ei, eix;	/* Result of integer expression */
  char * es, * esx;	/* Result of string expr */
  list l;
  struct tree_t cnode;	/* Temporary case node */
  struct tree_t sw;	/* Temp switch result */
  long tstack;
  record * r;
  var * v, * cv;
  object * o, * sot;
  field * f;
  long nobj, wasnobj;
  int ssize;	/* Local sort object size */
  
  if (!t)
    return 0;
  
  tstack = mark_tempstack();
  
  switch (t->type)
  {
  case T_INT:
    if (raw)
      fprintf(outfile, "%d", t->u.intval);
    else
      filter_int(t->u.intval);
    break;
  case T_STRING:
    if (raw)
      fputs(t->u.strval, outfile);
    else
      filter(t->u.strval);
    break;
  case T_EXPR:
    /* Open expression.  Evaluate result and send to current outfile */
    er = eval_expr(&ei, &es, t->u.expr.left);
    if (er == E_ERROR)
    {
      ret = 1;
      break;
    }
    else
    {
      static int prevline = -1;
      static char * prevfile = "";
      
      if (srcident && t->u.expr.oper == T_CRLF)
      {
        int thisline = t->lineno;
        char * thisfile = t->filename;
        
        if (strcmp(thisfile, prevfile) || thisline != prevline + 1)
        {
	  if (raw)
	    fprintf(outfile, "#line %d %s\n", thisline, thisfile);
	  else
	  {
	    filter("#line ");
	    filter_int(thisline);
	    filter(" ");
	    filter(thisfile);
	    filter("\n");
	  }
	}
	prevline = thisline;
	prevfile = thisfile;
      }
      if (er == E_INT)
	if (raw)
	  fprintf(outfile, "%d", ei);
	else
	  filter_int(ei);
      else
      {
	if (raw)
	  fputs(es, outfile);
	else
	  filter(es);
	/* The following is not entirely satisfactory... If there is a CR in
	   the string, and we are doing srcidents, then we need to output
	   another #line to reset the file position, otherwise it will be
	   advanced too far by the CR(s) in the string */
	if (srcident && strchr(es, '\n'))
	{
	  prevline = -1;	/* Force #line output next line */
	}
      }
      if (t->u.expr.oper == T_CRLF)
	if (raw)
	  fputc('\n', outfile);
	else
	  filter("\n");
    }
    break;
  case T_SET:
    er = eval_expr(&ei, &es, t->u.expr.right);
    if (er == E_ERROR)
    {
      ret = 1;
      break;
    }
    if (t->u.expr.left->type == T_INDEX)
    {
      erx = eval_expr(&eix, &esx, t->u.expr.left->u.expr.right);
      if (erx == E_ERROR)
      {
        ret = 1;
        break;
      }
      assign_array(t->u.expr.left->u.expr.left, erx, eix, esx, er, ei, es);
    }
    else
      assign_var(t->u.expr.left, er, ei, es);
    break;
  case T_IF:
    for (l = t->u.cond.conds; l; l = l->next)
    {
      cond_node * cn = &l->val->u.cn;
      if (!cn->cond)
        ret = execute(cn->stmts);  /* Must be ELSE clause (no cond) */
      else
      {
        er = eval_expr(&ei, &es, cn->cond);
        if (er == E_ERROR)
          ret = 1;
        else if (er == E_INT && ei)
          { ret = execute(cn->stmts); break; }
        else if (er == E_STRING && *es)
          { ret = execute(cn->stmts); break; }
      }
      if (ret)
        break;
    }
    break;
  case T_SWITCH:
    er = eval_expr(&ei, &es, t->u.sw.expr);
    if (er == E_ERROR)
    {
      ret = 1;
      break;
    }
    cnode.type = T_BINARY;
    cnode.u.expr.left = &sw;
    if (er == E_INT)
    {
      sw.type = T_INT;
      sw.u.intval = ei;
    }
    else
    {
      sw.type = T_STRING;
      sw.u.strval = es;
    }
    for (l = t->u.sw.conds; l; l = l->next)
    {
      cond_node * cn = &l->val->u.cn;
      if (!cn->cond)
        ret = execute(cn->stmts);  /* Must be default clause (no cond) */
      else
      {
        cnode.u.expr.right = cn->cond;
        cnode.u.expr.oper = cn->oper;
        er = eval_expr(&ei, &es, &cnode);
        if (er == E_ERROR)
          ret = 1;
        if (ei)
        { ret = execute(cn->stmts); break; }
      }
      if (ret)
        break;
    }
    break;
  case T_FOR:
    if ((ret = execute(t->u.loop.inits)))
      break;
    for (;;)
    {
      er = eval_expr(&ei, &es, t->u.loop.test);
      if (er != E_INT)
      {
        ret = 1;
        break;
      }
      if (!ei ||
          (ret = execute(t->u.loop.stmts)) ||
          (ret = execute(t->u.loop.assigns)))
        break;
    }
    break;
  case T_FORALL:
    wasnobj = 0;
    if (t->u.forall.datatype->type == T_VAR)
    {
      char duploop;
      
      r = (record *)lookup_hash(recht,
                    t->u.forall.datatype->u.var.varname);
      if (r)
        duploop = 0;
      else
      {
        /* datatype was not known record, look for outer loop ctl var */
        cv = existing_var(&t->u.forall.datatype->u.var);
        if (cv && cv->flags & VAR_ALIAS) cv = cv->rvar;
        if (cv && (cv->flags & VAR_FORALL))
        {
          r = cv->rec;
          duploop = 1;
        }
      }
      if (!r)
      {
        error(t, "unknown record type %s", t->u.forall.datatype->u.var.varname);
        ret = 1;
        break;
      }
      if (duploop && !r->chainindex)
      {
        error(t, "nested FORALL, but record type %s is not keyed",
               t->u.forall.datatype->u.var.varname);
        ret = 1;
        break;
      }
      v = create_var(&t->u.forall.var->u.var);
      if (!v)
      {
	error(t, "forall loop control variable %s already exists",
	       t->u.forall.var->u.var.varname);
	ret = 1;
	break;
      }
      init_forallvar(v, r);
      
      if (duploop)
        for (o = cv->obj, nobj = 0;
             o;
             o = (object *)o->data[r->chainindex], nobj++);
      else
        for (o = r->objs, nobj = 0; o; o = o->next, nobj++);
      
      sot = setup_sort(t->u.forall.sort, v, nobj);
      ssize = sortobjsize;
      if (sot)
      {
        object * s;
        if (duploop)
          for (o = cv->obj, s = sot;
               o;
               o = (object *)o->data[r->chainindex], 
               s = (object*)((char *)s+ssize))
            s->next = o;
        else
          for (o = r->objs, s = sot; o; o = o->next,
               s = (object*)((char *)s+ssize))
            s->next = o;
        keep(sot, &nobj, ssize, t->u.forall.keep, v, nsort,
                (int (*)(const void *, const void *))sort_comp);
        wasnobj = nobj;
        v->flags |= VAR_FIRST;
        v->flags &= ~VAR_LAST;
        /* We go backwards, because the default ordering is reverse anyway
           because of the way linked lists are created by adding to head. */
        for (s = (object *)((char *)sot + ssize * (nobj-1));
             nobj;
             nobj--, s = (object *)((char *)s - ssize))
	{
	  o = s->next;
	  set_forallvar(v, o);
	  if (nobj == 1)
	    v->flags |= VAR_LAST;
	  ret = execute(t->u.forall.stmts);
	  v->flags &= ~VAR_FIRST;
	  if (ret)
	    break;
	}
        free(sot);
      }
      unhash_destroy_var(v);
    }
    else
    {
      /* FORALL datatype is in form  r.f  where r is a control variable
         for an enclosing FORALL, and f is the name of a field.  The fields
         of objects represented by r are searched for a matching field called
         f.  First, the reference lists are searched.  If the field is not
         found, the normal data fields are searched.  If a match is found
         in the reference list, then all objects in that list are iterated
         in the body of the loop.  If f is a normal data field, then all
         instances of objects of type r are iterated (i.e. all objects of
         type r with the current definitive field (key)). */
      method_node * mn = &t->u.forall.datatype->u.meth;
      object_list * l;
      
      if (mn->obj->type != T_VAR)
      { error(t, "more than one '.' in datatype specification"); ret=1; break; }
      cv = existing_var(&mn->obj->u.var);
      if (cv && cv->flags & VAR_ALIAS) cv = cv->rvar;
      if (!cv || !(cv->flags & VAR_FORALL))
      { error(t, "datatype %s is not a FORALL loop control var",
               mn->obj->u.var.varname); 
        ret=1; break; }
      v = create_var(&t->u.forall.var->u.var);
      if (!v)
      { error(t, "forall loop control variable %s already exists",
               t->u.forall.var->u.var.varname); 
        ret=1; break; }
      
      f = init_nested_forallvar(v, cv, mn->cust);
      if (!f)
      { error(t, "datatype %s has no such referring record type",
               mn->obj->u.var.varname); 
        ret=1; break; }
      for (l = (object_list *)cv->obj->data[f->rb->objindex], nobj = 0;
           l; 
           l = l->next, nobj++);
      sot = setup_sort(t->u.forall.sort, v, nobj);
      ssize = sortobjsize;
      if (sot)
      {
        object * s;
        for (l = (object_list *)cv->obj->data[f->rb->objindex], s = sot;
             l; 
             l = l->next, s = (object*)((char *)s+ssize))
            s->next = l->obj;
        keep(sot, &nobj, ssize, t->u.forall.keep, v, nsort,
                (int (*)(const void *, const void *))sort_comp);
        /*if (nsort)
          qsort(sot, nobj, ssize, 
                (int (*)(const void *, const void *))sort_comp);*/
        wasnobj = nobj;
        v->flags |= VAR_FIRST;
        v->flags &= ~VAR_LAST;
        /* We go backwards, because the default ordering is reverse anyway
           because of the way linked lists are created by adding to head. */
        for (s = (object *)((char *)sot + ssize * (nobj-1));
             nobj;
             nobj--, s = (object *)((char *)s - ssize))
	{
	  o = s->next;
	  set_forallvar(v, o);
	  if (nobj == 1)
	    v->flags |= VAR_LAST;
	  ret = execute(t->u.forall.stmts);
	  v->flags &= ~VAR_FIRST;
	  if (ret)
	    break;
	}
        free(sot);
      }
      unhash_destroy_var(v);
    }
    if (!wasnobj && t->u.forall.ifnone)
      ret = execute(t->u.forall.ifnone);
    break;
  case T_OUTFILE:
    flush_filters();
    delete_filters();
    for (l = t->u.l; l; l = l->next)
      /* Iterate through list of T_METHOD (with null thisvars) */
      append_filter(l->val);
    break;
  case T_FUNCTION:
    break;
  default:
    /* Shouldn't happen */
    break;
  }

  /* Must return thru here */
  pop_tempstack(tstack);
  
  return ret;
}

int execute(list root)
{
  while (root)
  {
    if (exec_stmt(root->val))
      return 4;
    root = root->next;
  }
  return 0;
}

