#if !defined(DGLEX_H_Incl)
#define DGLEX_H_Incl
/*****************************************************************************
*   $Id: dglex.h,v 1.2 2000/06/29 13:13:13 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    techno@spirit.net.au
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: dglex.h,v $
* Revision 1.2  2000/06/29 13:13:13  tekno
* Added -c option to allow commenting out of -m macros
*
* Revision 1.1.1.1  2000/06/22 15:38:49  tekno
* Opened source
*
*
*****************************************************************************/

#include <stdio.h>

typedef enum node_type_t
{
  T_SET
 ,T_EXPR	/* Output (open) expression */
 ,T_PARM	/* Formal parm with opt initialiser */
 ,T_IF
 ,T_COND
 ,T_SWITCH
 ,T_FOR
 ,T_FORALL
 ,T_FUNCTION	/* Method definition */
 ,T_OUTFILE
 ,T_CRLF
 ,T_INT
 ,T_FLOAT
 ,T_STRING
 ,T_VAR
 ,T_METHOD	/* Method invocation */
 ,T_UNARY
 ,T_BINARY
 ,T_REGEXP	/* Compiled regular expression */
 ,T_INDEX       /* Array indexing */
 ,T_REFERS      /* Formal parameter pass by reference */
} node_type;


typedef enum method_t
{
#define DG_METHOD(tag, type, ret, funcret, func, funcsig, nparms, name, desc)\
  tag,
#include "dgmethods.h"
#undef DG_METHOD
  M_CUSTOM		/* Custom method */
} method;


typedef struct list_t
{
  struct list_t * next;
  struct tree_t * val;
} * list;
    

typedef struct expr_node_t
{
  /* Binary/unary expression parse tree node.  'left' is null for unary. 
     'oper' may also be SET for an assignment (in which case 'left' will
     be the VARNAME node).  'right' may be an index expression
     applied to the array 'left'.  For T_PARM, left is a T_VAR and right
     is an optional initialiser expression. */
  int oper;
  struct tree_t * right, * left;
} expr_node;

typedef struct var_node_t
{
  /* Named reference to a variable. */
  char * varname;	/* Variable name, or null if 'this'. */
  struct var_t * var;	/* Variable object, or null if always lookup.
  			   May be non-null for global variables which
  			   will always exist in the same location. */
  struct var_node_t * next;	/* Next var_node whose var pointer points
  				   to the same var. */
  char global;		/* True if variable had global scope qualifier '::' */
  char alias;		/* True if this is an alias i.e. varname is different
  			   but var points to shared variable instance.  Used
  			   for pass-by-reference formal parameters. */
} var_node;

typedef struct method_node_t
{
  /* Method invocation node */
  struct tree_t * obj;	/* 'this' object */
  method name;		/* Method number (translated from VARNAME string) */
  list parms;		/* Parameter list, or null */
  char * cust;		/* Custom method name */
} method_node;

typedef struct cond_node_t
{
  int oper;		/* Comparison operator, or 0 if cond is a boolean.
			   If not 0, a boolean expression is created as
			     '$sw oper cond'
			   where $sw is the value of the SWITCH expression. */
  struct tree_t * cond;	/* Condition of an if, elseif, else or case.  For
  			   an else, this will be null. */
  list stmts;		/* Statements to execute when cond true (or null) */
} cond_node;

typedef struct cond_stmt_t
{
  /* Used for IF statements */
  unsigned flags;	/* Optimisation flags */
  list conds;		/* List of cond_node's */
} cond_stmt;

typedef struct switch_stmt_t
{
  /* Used for SWITCH statements */
  unsigned flags;	/* Optimisation flags */
  list conds;		/* List of cond_node's */
  struct tree_t * expr;	/* Switch expression */
} switch_stmt;

typedef struct forall_stmt_t
{
  struct tree_t * datatype;	/* Datatype to enumerate.  May be var_node
  				   for simple datatype, or method_node for
  				   var.list syntax. */
  struct tree_t * var;	/* Loop control variable (T_VAR node) */
  struct tree_t * keep; /* Optional subsetting statement */
  list sort;		/* Sort parm list, or null */
  list ifnone;		/* ifnone clause */
  list stmts;		/* Body, or null */
} forall_stmt;

typedef struct for_stmt_t
{
  /* Any of these may be null except test (unlike C) */
  list inits;		/* Initialisation */
  struct tree_t * test;	/* Test expression */
  list assigns;		/* Assignments */
  list stmts;		/* Body */
} for_stmt;

typedef struct method_def_t
{
  struct tree_t * thistype;	/* INT, STRING or VAR expression node */
  char * name;			/* Method name */
  list parms;			/* Parameter list, or null */
  list stmts;			/* Statement list, or null */
  struct tree_t * ret;		/* Return expression */
} method_def;

typedef struct tree_t
{
  node_type type;
  int lineno;
  char * filename;
  union
  {
    int intval;		/* T_INT */
    char * strval;	/* T_STRING */
    list l;		/* T_OUTFILE */
    var_node var;	/* T_VAR */
    expr_node expr;	/* T_EXPR, T_SET, T_REF, subexpressions, T_PARM */
    method_node meth;	/* method invocations */
    for_stmt loop;	/* T_FOR */
    forall_stmt forall;	/* T_FORALL */
    cond_stmt cond;	/* T_IF */
    switch_stmt sw;	/* T_SWITCH */
    cond_node cn;	/* T_COND */
    method_def mdef;	/* T_FUNCTION */
    struct re_pattern_buffer *rx;  /* T_REGEXP (i.e. regex_t).  This type
    					of node is a mutation of an expression
    					which reduces to a constant string
    					used as the RHS of a ~ or ~= operator.
    				   */
  } u;
} * tree;

typedef enum expr_result_t
{
  E_INT,
  E_STRING,
  E_VAR,
  E_ERROR
} expr_result;

/* Globals for communication between lex and yacc (instantiated in dgscan.c) */
extern FILE * yyin;
extern int yyleng;
extern char * srcfilename;
extern int lex_line;	/* Approx token end line */
extern int lex_tkstart;	/* Token start line */
extern int pass1error;
extern int verbose;	/* -v */
extern int raw;		/* -r */
extern int preparse;	/* -p */
extern int lbl;		/* -l */
extern int srcident;	/* -# */
extern int maxsrcfiles;	/* -i n (default 255) */
extern int maxsrcnest;	/* -j n (default 10) */
extern char * cont_str; /* -c xxx */
extern char * macro_str; /* -m xxx */

#define MAX_INCDIRS     100
extern int nincdirs;
extern char * incdirs[MAX_INCDIRS];	/* -I dir */

#define MAX_COMMENT_STRS 4
extern int ncmntstrs;
extern char * cmntstrs[MAX_COMMENT_STRS];/* -c str */

extern int xdebug;
extern list tmpl_root;
extern list mdef_root;
extern FILE * outfile;	/* Current output file */

/* Lexer functions */
int yylex(void);
int init_lex(int free_string, int skip_hashbang);

/* Error handler */
expr_result error(tree t, char * msg, ...);


/* Control character conversions: platform dependent */
#define CTRL_0 1
#define CTRL_1 2
#define CTRL_2 3
#define CTRL_3 4
#define CTRL_4 5
#define CTRL_5 6
#define CTRL_6 14
#define CTRL_7 15
#define CTRL_8 16
#define CTRL_9 17


#endif     /* PGHlex_h_Incl */
