%{
/*****************************************************************************
*   $Id: dgparse.y,v 1.1.1.1 2000/06/22 15:39:03 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    techno@spirit.net.au
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: dgparse.y,v $
* Revision 1.1.1.1  2000/06/22 15:39:03  tekno
* Opened source
*
*
*****************************************************************************/
#include "dglex.h"
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
%}

/* S/R conflicts to expect.  One only, caused by ambiguity in syntax: we
   allow an implicit concatenation operator so is
     alpha . beta ( foo )
   parsed as
     alpha.beta(foo)           - i.e. method invocation with parameters
   or
     alpha.beta <concat> (foo) - i.e. method with no parms, concat parenthesised
                                 expression.
   The default rule (shift) ensures the former interpretation.
*/
%expect 1

%union
{
  int lex_int;       /* For INTEGER */
  char * lex_str;    /* STRING.
                        malloc()'d in lex */
  double lex_float;  /* For FLOATING */
  
  /* Now stuff for non-terminal nodes */
  tree   lex_tree;		/* Generic parse node */
  list   lex_list;		/* Generic list of nodes */
}
 
%start tmpl
%token <lex_int> INTEGER
%token <lex_float> FLOATING
%token <lex_str> STRING
%token <lex_str> VAR				/* variable identifier */
%token IF
%token ELSE
%token ELSEIF
%token ENDIF
%token FOR
%token FORALL
%token NEXT
%token KEEP
%token IFNONE
%token SORT
%token SWITCH
%token CASE
%token ENDSWITCH
%token METHOD
%token RETURN
%token THIS
%token GLOBAL
%token OUTFILE
%token GE
%token LE
%token NE
%token SHL
%token SHR
%token LOGAND
%token LOGOR
%token MATCHES
%token STRICTMATCHES
%token DBLSEMI
%token REFERS

%type <lex_list> tmpl opt_statements statements opt_elseifs elseifs
%type <lex_list> opt_cases cases opt_parms parms assignments opt_assignments
%type <lex_list> opt_formal_parms formal_parms opt_filters filters

%type <lex_tree> tmpl_statement opt_else elseif case brkt_expr expression
%type <lex_tree> primary unary_expr primarycat assignment datatype thistype
%type <lex_tree> method_def opt_subset opt_parm_init formal_parm filter


%type <lex_list> opt_sort opt_ifnone
%type <lex_int> unary_op multiplicative_op additive_op 
%type <lex_int> shift_op comparison_op equality_op

%left <lex_int> ';' DBLSEMI
%left <lex_int> ','
%left <lex_int> LOGOR
%left <lex_int> LOGAND
%left <lex_int> '|'
%left <lex_int> '^'
%left <lex_int> '&'
%left <lex_int> EQ NE MATCHES STRICTMATCHES
%left <lex_int> '<' '>' LE GE
%left <lex_int> SHL SHR
%left <lex_int> '+' '-'
%left <lex_int> '*' '/' '%'
%right <lex_int> unary
%left <lex_int> '.' '('

 
%{

list tmpl_root;	/* Set to list of statements comprising entire template. */
int nest;	/* Statement nesting level */

#define MAX_NESTING 8	/* Levels we are interested in reporting */
static struct
{
  int lineno;
  char * filename;
  node_type nt;
} sstack[MAX_NESTING];

static void spush(node_type nt)
{
  /* Helps with error reporting.  Remembers file/line of start of compound
     statement which is being parsed.  If EOF encountered, then if any
     statement has not been terminated properly (e.g. forgetting an endif
     in an if statement), then we can indicate the line at which the
     statement started. */
  if (nest < MAX_NESTING)
  {
    /*fprintf(stderr, "pushd nest %d\n", nest);*/
    sstack[nest].lineno = lex_line;
    sstack[nest].filename = srcfilename;
    sstack[nest].nt = nt;
  }
}

static list start_list(tree t)
{
  list l;
  
  if (!t)  /* because called for immediate execution */
    return 0;
  l = (list)malloc(sizeof(*l));
  l->next = 0;
  l->val = t;
  if (!t)
    yyerror("start_list: null tree");
    
  return l;
}

static list add_tree(list l, tree t)
{
  /* Add t to end of list l. */
  list r;
  if (!l || !t)  /* because called for immediate execution */
    return l;
  for (r = l; r->next; r = r->next);
  r->next = start_list(t);
  return l;
}

static tree new_tree(node_type nt)
{
  /* We could be smarter about allocating the minimum size, but for
     now just allocate max size */
  tree t = (tree)calloc(sizeof(*t), 1);
  t->type = nt;
  t->lineno = lex_line;
  t->filename = srcfilename;
  return t;
}

static tree make_condition(int op, tree expr, list stmts)
{
  tree t = new_tree(T_COND);
  t->u.cn.oper = op;
  t->u.cn.cond = expr;
  t->u.cn.stmts = stmts;
  return t;
}

static tree make_exprnode(node_type nt, ...)
{
  tree t;	/* Return value */
  va_list ap;
  
  va_start(ap, nt);
  t = new_tree(nt);
  switch (nt)
  {
  case T_INT:
    t->u.intval = va_arg(ap, int);
    break;
  case T_FLOAT:
    /* Reserved for future use.  For now, floats are converted to int. */
    t->u.intval = (int)va_arg(ap, double);
    t->type = T_INT;
    break;
  case T_STRING:
    t->u.strval = va_arg(ap, char *);
    break;
  case T_VAR:
    t->u.var.varname = va_arg(ap, char *);
    t->u.var.global = va_arg(ap, int);
    break;
  case T_PARM:
    t->u.expr.left = make_exprnode(T_VAR, va_arg(ap, char *), 0);
    t->u.expr.right = va_arg(ap, tree);
    break;
  case T_REFERS:
    t->u.expr.left = make_exprnode(T_VAR, va_arg(ap, char *), 0);
    break;
  case T_METHOD:
    t->u.meth.obj = va_arg(ap, tree);  /* May be null for filter */
    t->u.meth.name = translate_methodname(t->u.meth.cust = va_arg(ap, char *));
    t->u.meth.parms = va_arg(ap, list);
    break;
  case T_UNARY:
    t->u.expr.oper = va_arg(ap, int);
    t->u.expr.right = va_arg(ap, tree);
    break;
  case T_BINARY:
    t->u.expr.left = va_arg(ap, tree);
    t->u.expr.oper = va_arg(ap, int);
    t->u.expr.right = va_arg(ap, tree);
    break;
  case T_INDEX:
    t->u.expr.left = va_arg(ap, tree);
    t->u.expr.right = va_arg(ap, tree);
    break;
  }
  va_end(ap);
  return t;
}

static tree make_statement(node_type nt, ...)
{
  tree t;	/* Return value */
  tree u, v, w;
  list l;
  va_list ap;
  int g;
  char * vn;
  
  va_start(ap, nt);
  t = new_tree(nt);
  switch (nt)
  {
  case T_SET:	/* Parms: varname, index, expression, global. */
    t->u.expr.oper = nt;
    vn = va_arg(ap, char *);
    u = va_arg(ap, tree);  /* Index expression */
    w = va_arg(ap, tree);  /* Value expr */
    g = va_arg(ap, int);   /* Global var */
    if (u)
      t->u.expr.left = make_exprnode(T_INDEX, make_exprnode(T_VAR, vn, g), u);
    else
      t->u.expr.left = make_exprnode(T_VAR, vn, g);
    t->u.expr.right = w;
    break;
  case T_IF:	/* Parms: expression, list of stmts for IF, list of cond nodes
  			for elseifs, single cond node for else. */
    /* Make cond for IF (cond) stmts part */
    v = va_arg(ap, tree);
    u = make_condition(0, v, va_arg(ap, list));
    t->u.cond.conds = start_list(u);
    l = va_arg(ap, list);
    while (l)
    {
      v = l->val; 
      add_tree(t->u.cond.conds, v);
      l = l->next;
    }
    v = va_arg(ap, tree);
    if (v)
      add_tree(t->u.cond.conds, v);
    break;
  case T_SWITCH:	/* Parms: expression, list of cond nodes for cases. */
    t->u.sw.expr = va_arg(ap, tree);
    t->u.sw.conds = va_arg(ap, list);
    break;
  case T_FOR:	/* Parms: list of init, test expr, list of assign, body list */
    t->u.loop.inits = va_arg(ap, list);
    t->u.loop.test = va_arg(ap, tree);
    t->u.loop.assigns = va_arg(ap, list);
    t->u.loop.stmts = va_arg(ap, list);
    break;
  case T_FORALL:	/* Parms: char * for datatype, 
                        char * for var,
                        keep expression or null (tree),
  			list of sort expressions,
  			list of ifnone statements,
  			list of statements */
    t->u.forall.datatype = va_arg(ap, tree);
    t->u.forall.var = make_exprnode(T_VAR, va_arg(ap, char *), 0);
    t->u.forall.keep = va_arg(ap, tree);
    t->u.forall.sort = va_arg(ap, list);
    t->u.forall.ifnone = va_arg(ap, list);
    t->u.forall.stmts = va_arg(ap, list);
    break;
  case T_FUNCTION:	/* Parms: tree thistype, char * for method name,
  			list of formal parms, list of statements,
  			return expression. */
    t->u.mdef.thistype = va_arg(ap, tree);
    t->u.mdef.name = va_arg(ap, char *);
    t->u.mdef.parms = va_arg(ap, list);
    t->u.mdef.stmts = va_arg(ap, list);
    t->u.mdef.ret = va_arg(ap, tree);
    break;
  case T_EXPR:	/* Parms: expression (tree), add CRLF (int) */
    t->u.expr.left = va_arg(ap, tree);
    t->u.expr.oper = va_arg(ap, int) ? T_CRLF : T_EXPR;
    break;
  case T_OUTFILE:  /* Parms: filter list */
    t->u.l = va_arg(ap, list);
    break;
  default:
    break;
  }
  va_end(ap);
  return t;
}

static tree imm_execute(tree t)
{
  /* Called at end of parsing outermost level statement when interpreting
     in immediate mode i.e. no -p (preparse) option.  Execute the statement
     then discard the tree of statements t to free resources.  Naturally,
     this mode of execution requires methods to be defined before they
     are invoked. */
  int rc;
  extern int exec_stmt(tree t);
  extern int recycle_stmt(tree t);
  
  if (!t)
    return 0;	/* This happens if syntax error */
  
  if (t->type == T_FUNCTION)
  {
    grok_mdef(t);
    return t;  /* Method defs must be retained */
  }
  rc = exec_stmt(t);
  recycle_stmt(t);
  if (rc)
    fprintf(stderr, "%s:%d: Error executing statement, rc=%d.\n", 
      t->filename, t->lineno, rc);
  return 0;
}
 
%}
 
%%
tmpl:
	opt_statements
		{ 
		  tmpl_root = $1;
		  if (!preparse)
		  {
		    extern void flush_filters(void);
		    flush_filters();
		  }
		}
	;

method_def:
	method_head thistype '.' VAR '(' opt_formal_parms ')'
	opt_statements
	RETURN brkt_expr
		{ $$ = make_statement(T_FUNCTION, $2, $4, $6, $8, $10); 
		}
	|
	method_head error RETURN brkt_expr
		{ $$ = 0;
		  yyerror("method definition error: expecting subject type"); }
	|
	method_head thistype error RETURN brkt_expr
		{ $$ = 0;
		  yyerror("method definition error: expecting '.'"); }
	|
	method_head thistype '.' error RETURN brkt_expr
		{ $$ = 0; 
		  yyerror("method definition error: expecting method name"); }
	|
	method_head thistype '.' VAR error RETURN brkt_expr
		{ $$ = 0; 
		  yyerror("method definition error: expecting formal parms"); }
	|
	method_head thistype '.' VAR '(' opt_formal_parms ')'
        error RETURN brkt_expr
		{ $$ = 0; 
		  yyerror("method definition error: error in body"); }
	;

method_head:
	METHOD
		{ spush(T_FUNCTION); }
	;

if_head:
	IF
		/*{ spush(T_IF); }*/
	;

switch_head:
	SWITCH
		{ spush(T_SWITCH); }
	;

for_head:
	FOR
		{ spush(T_FOR); }
	;

forall_head:
	FORALL
		{ spush(T_FORALL); }
	;

tmpl_statement:
	expression ';'
		{ $$ = make_statement(T_EXPR, $1, 1); }
	|
	expression DBLSEMI
		{ $$ = make_statement(T_EXPR, $1, 0); }
	|
	assignment ';'
	|
	if_head brkt_expr
	opt_statements
	opt_elseifs
	opt_else
	ENDIF
		{ $$ = make_statement(T_IF, $2, $3, $4, $5); }
	|
	if_head error ENDIF
		{ $$ = 0;
		  yyerror("if statement error: expecting control expression"); }
	|
	if_head brkt_expr error ENDIF
		{ $$ = 0;
		  yyerror("if statement: error in body"); }
	|
	switch_head brkt_expr
	opt_cases
	ENDSWITCH
		{ $$ = make_statement(T_SWITCH, $2, $3); }
	|
	switch_head error ENDSWITCH
		{ $$ = 0;
		  yyerror("switch statement: expecting control expression"); }
	|
	switch_head brkt_expr error ENDSWITCH
		{ $$ = 0;
		  yyerror("switch statement: error in body"); }
	|
	forall_head datatype VAR opt_subset opt_sort
	opt_ifnone
	opt_statements
	NEXT
		{ $$ = make_statement(T_FORALL, $2, $3, $4, $5, $6, $7); }
	|
	forall_head error NEXT
		{ $$ = 0;
		  yyerror("forall statement: expecting datatype"); }
	|
	forall_head datatype error NEXT
		{ $$ = 0;
		  yyerror("forall statement: expecting control variable"); }
	|
	forall_head datatype VAR error NEXT
		{ $$ = 0;
		  yyerror("forall statement: expecting options/statements"); }
	|
	for_head '(' opt_assignments ';' expression ';' opt_assignments ')'
	opt_statements
	NEXT
		{ $$ = make_statement(T_FOR, $3, $5, $7, $9); }
	|
	for_head error NEXT
		{ $$ = 0;
		  yyerror("for statement: expecting '('"); }
	|
	for_head '(' error ')' opt_statements NEXT
		{ $$ = 0;
		  yyerror("for statement: error in loop control"); }
	|
	OUTFILE opt_filters ';'
		{ $$ = make_statement(T_OUTFILE, $2); }
	|
	OUTFILE error ';'
		{ $$ = 0;
		  yyerror("outfile statement error"); }
	|
	method_def
	  /* Note that method defs are invisible unless declared at the
	     outermost level i.e. not in an IF, FOR statement etc. */
	;

opt_statements:
	/* empty */
		{ $$ = 0; }
	|
	statements
	;

statements:
	{ nest++; spush(T_EXPR); } tmpl_statement
		{ nest--; 
		  if (!preparse && !nest)
		    $$ = start_list(imm_execute($2));
		  else
		    $$ = start_list($2);
		}
	|
	statements {nest++; spush(T_EXPR); } tmpl_statement
		{ nest--; 
		  if (!preparse && !nest)
		    $$ = add_tree($1, imm_execute($3));
		  else
		    $$ = add_tree($1, $3); 
		}
	;

opt_ifnone:
	/* empty */
		{ $$ = 0; }
	|
	IFNONE statements ENDIF
		{ $$ = $2; }
	|
	IFNONE error ENDIF
		{ $$ = 0;
		  yyerror("ifnone clause error"); }
	;

opt_assignments:
	/* empty */
		{ $$ = 0; }
	|
	assignments
	;

assignments:
	assignment
		{ $$ = start_list($1); }
	|
	assignments ',' assignment
		{ $$ = add_tree($1, $3); }
	;

assignment:
	VAR '=' expression
		{ $$ = make_statement(T_SET, $1, 0, $3, 0); }
	|
	VAR '[' expression ']' '=' expression
		{ $$ = make_statement(T_SET, $1, $3, $6, 0); }
	|
	GLOBAL VAR '=' expression
		{ $$ = make_statement(T_SET, $2, 0, $4, 1); }
	|
	GLOBAL VAR '[' expression ']' '=' expression
		{ $$ = make_statement(T_SET, $2, $4, $7, 1); }
	|
	THIS '=' expression
		{ $$ = make_statement(T_SET, 0, 0, $3, 0); }
	;

opt_elseifs:
	/* empty */
		{ $$ = 0; }
	|
	elseifs
	;

elseifs:
	elseif
		{ $$ = start_list($1); }
	|
	elseifs elseif
		{ $$ = add_tree($1, $2); }
	;

elseif:
	ELSEIF brkt_expr opt_statements
		{ $$ = make_condition(0, $2, $3); }
	;

opt_else:
	/* empty */
		{ $$ = 0; }
	|
	ELSE opt_statements
		{ $$ = make_condition(0, 0, $2); }
	;

opt_cases:
	/* empty */
		{ $$ = 0; }
	|
	cases
	;

cases:
	case
		{ $$ = start_list($1); }
	|
	cases case
		{ $$ = add_tree($1, $2); }
	;

case:
	CASE brkt_expr opt_statements
		{ $$ = make_condition(EQ, $2, $3); }
	|
	CASE comparison_op brkt_expr opt_statements
		{ $$ = make_condition($2, $3, $4); }
	|
	CASE equality_op brkt_expr opt_statements
		{ $$ = make_condition($2, $3, $4); }
	;

opt_filters:
	/* empty */
		{ $$ = 0; }
	|
	filters
	;

filters:
	filter
		{ $$ = start_list($1); }
	|
	filters filter
		{ $$ = add_tree($1, $2); }
	;

filter:
	VAR opt_parms
		{ $$ = make_exprnode(T_METHOD, 0, $1, $2); }
	;

brkt_expr:
	'(' expression ')'
		{ $$ = $2; }
	;

primary:
	thistype
	|
	GLOBAL VAR
		{ $$ = make_exprnode(T_VAR, $2, 1); }
	|
	GLOBAL VAR '[' expression ']'
		{ $$ = make_exprnode(T_INDEX, 
		                     make_exprnode(T_VAR, $2, 1), $4); }
	|
	THIS
		{ $$ = make_exprnode(T_VAR, 0, 0); }
	|
	brkt_expr
	|
	primary '.' VAR opt_parms
		{ $$ = make_exprnode(T_METHOD, $1, $3, $4); }
	;

expression:
	unary_expr
	|
	expression multiplicative_op expression		%prec '*'
		{ $$ = make_exprnode(T_BINARY, $1, $2, $3); }
	|
	expression additive_op expression		%prec '+'
		{ $$ = make_exprnode(T_BINARY, $1, $2, $3); }
	|
	expression shift_op expression			%prec SHL
		{ $$ = make_exprnode(T_BINARY, $1, $2, $3); }
	|
	expression comparison_op expression		%prec '<'
		{ $$ = make_exprnode(T_BINARY, $1, $2, $3); }
	|
	expression equality_op expression		%prec EQ
		{ $$ = make_exprnode(T_BINARY, $1, $2, $3); }
	|
	expression '&' expression
		{ $$ = make_exprnode(T_BINARY, $1, '&', $3); }
	|
	expression '^' expression
		{ $$ = make_exprnode(T_BINARY, $1, '^', $3); }
	|
	expression '|' expression
		{ $$ = make_exprnode(T_BINARY, $1, '|', $3); }
	|
	expression LOGAND expression
		{ $$ = make_exprnode(T_BINARY, $1, LOGAND, $3); }
	|
	expression LOGOR expression
		{ $$ = make_exprnode(T_BINARY, $1, LOGOR, $3); }
	;


unary_expr:
	primarycat		%prec unary
	|
	unary_op unary_expr	%prec unary
		{ $$ = make_exprnode(T_UNARY, $1, $2); }
	;

primarycat:
	primary
	|
	primarycat primary
		{ $$ = make_exprnode(T_BINARY, $1, '#', $2); }
	;

opt_parms:
	/* empty */
		{ $$ = 0; }
	|
	'(' parms ')'
		{ $$ = $2; }
	;

parms:
	expression
		{ $$ = start_list($1); }
	|
	parms ',' expression
		{ $$ = add_tree($1, $3); }
	;

opt_formal_parms:
	/* empty */
		{ $$ = 0; }
	|
	formal_parms
	;

formal_parms:
	formal_parm
		{ $$ = start_list($1); }
	|
	formal_parms ',' formal_parm
		{ $$ = add_tree($1, $3); }
	;

formal_parm:
	VAR opt_parm_init
		{ $$ = make_exprnode(T_PARM, $1, $2); }
	|
	VAR REFERS
		{ $$ = make_exprnode(T_REFERS, $1); }
	;

opt_parm_init:
	/* empty */
		{ $$ = 0; }
	|
	'=' expression
		{ $$ = $2; }
	;

unary_op:
	'!' { $$ = '!'; }	/* Logical inversion */
	|
	'-' { $$ = '-'; }	/* Negation, string reversion for sort*/
	|
	MATCHES { $$ = '~'; }	/* Bitwise inversion (this operator in binary
				   form is used for string regexp matching) */
	;

multiplicative_op:
	'*' { $$ = '*'; }
	|
	'/' { $$ = '/'; }
	|
	'%' { $$ = '%'; }
	;

additive_op:
	'+' { $$ = '+'; }	/* Addition, string concat */
	|
	'-' { $$ = '-'; }
	;

shift_op:
	SHL { $$ = SHL; }	/* Bitwise shift */
	|
	SHR { $$ = SHR; }
	;

comparison_op:
	'<' { $$ = '<'; }
	|
	'>' { $$ = '>'; }
	|
	GE { $$ = GE; }		/* >= */
	|
	LE { $$ = LE; }		/* <= */
	;

equality_op:
	EQ { $$ = EQ; }		/* == */
	|
	NE { $$ = NE; }		/* != */
	|
	MATCHES	 { $$ = MATCHES; }	/* ~ regexp match anywhere */
	|
	STRICTMATCHES { $$ = STRICTMATCHES; }	
			/* ~= strict regexp match (anchor begin and end) */
	;

datatype:
	VAR
		{ $$ = make_exprnode(T_VAR, $1, 0); }
	|
	VAR '.' VAR
		{ $$ = make_exprnode(T_METHOD, 
		                     make_exprnode(T_VAR, $1, 0), $3, 0);
		}
	;

thistype:
	INTEGER
		{ $$ = make_exprnode(T_INT, $1); }
	|
	FLOATING
		{ $$ = make_exprnode(T_INT, (int)$1); }
	|
	STRING
		{ $$ = make_exprnode(T_STRING, $1); }
	|
	VAR
		{ $$ = make_exprnode(T_VAR, $1, 0); }
	|
	VAR '[' expression ']'
		{ $$ = make_exprnode(T_INDEX, 
		                     make_exprnode(T_VAR, $1, 0), $3); }
	;

opt_sort:
	/* empty */
		{ $$ = 0; }
	|
	SORT '(' parms ')'
		{ $$ = $3; }
	;

opt_subset:
	/* empty */
		{ $$ = 0; }
	|
	KEEP brkt_expr
		{ $$ = $2; }
	;

 
%%
 
yyerror(char * s) 
{
  node_type nt = sstack[nest].nt;
  
  fprintf(stderr, "%s:%d: %s\n", srcfilename,
        	lex_line, s);
  if (nt >= T_IF)
    fprintf(stderr, "...while parsing %s (nest %d) at %s:%d\n",
      nt == T_FUNCTION ? "method definition" :
      nt == T_IF ? "if statement" :
      nt == T_FOR ? "for statement" :
      nt == T_FORALL ? "forall statement" :
      "switch statement",
      nest,
      sstack[nest].filename,
      sstack[nest].lineno);

  pass1error = 1;
}
/*
yywrap(void)
{
  return(1);
}
*/
 
