/*****************************************************************************
*   $Id: dgscan.c,v 1.3 2002/02/17 21:50:23 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    techno@spirit.net.au
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: dgscan.c,v $
* Revision 1.3  2002/02/17 21:50:23  tekno
* Rebuild
*
* Revision 1.2  2000/06/29 13:13:13  tekno
* Added -c option to allow commenting out of -m macros
*
* Revision 1.1.1.1  2000/06/22 15:39:23  tekno
* Opened source
*
*
*****************************************************************************/
/*********************************************************************
** dgscan.c - hand-crafted version of flex-generated dglex.c.
*  This is done to ease portability problems and to hopefully
*  increase scanning speed for lexical tokens.  Note that string
*  tokens might be quite enormous since 'free text' is treated
*  like a string.
*
*  This scanner is smart enough to insert implied semicolons
*  between free text and following if, for, EOF tokens etc.
**********************************************************************/

#ifdef DEBUG_LEX
#define DEBUGPRINT(x) if (xdebug) printf x
#else 
#define DEBUGPRINT(x) /* nothing */
#endif                /* DEBUG_LEX */

/*
   Original lex character classes 
D	[0-9]
E	[DEde][-+]?{D}+
H	[0-9A-Fa-f]
A	[A-IJ-RS-Za-ij-rs-z_]
B	[A-IJ-RS-Za-ij-rs-z_0-9]
N	{A}{B}*

*/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "dglex.h"
#include "y.tab.h"

/*
 * Copy a few macros and stuff from lex output.
 */

#define U(x) x
#define YYNEWLINE '\n'
#define CLSBRACE '}'
#define OPNBRACE '{'
#define YYLMAX 256	/* Maximum token size, strings handled differently */
#define input() \
 (((yytchar=yysptr>yysbuf?U(*--yysptr):getc(yyin))==YYNEWLINE? \
    lex_line++,yytchar:yytchar)==EOF?0:yytchar)
#define unput(c) {yytchar= (c);*yysptr++=yytchar;if (c==YYNEWLINE)lex_line--;}
int yyleng; 		/* Length of token (incl. long strings) */
char yytext[YYLMAX];	/* Non-string token chars (null terminated) */

/* Strings are handled differently.  When reading a string token, the
   first YYLMAX-1 chars will be read into yytext as per normal.  If the
   string is longer than this, the file position of the _next_ char
   will be stored.  The file is scanned (without storing) until the
   end of string is found.  A buffer for the entire string will be allocated
   and the file contents re-read.  This is efficient for small strings,
   yet doesn't waste memory for very long strings. */
static fpos_t yynsc;	/* Pos of YYLMAXth char of current string. */
static char * yylstr;	/* Allocated storage for strings (long or short).
                           This is different storage for each string,
                           caller should free. */
static int bump_linect;

/* Stack for unput() */
char yysbuf[YYLMAX];
char *yysptr = yysbuf;
int yytchar;

FILE *yyin;
YYSTYPE yylval;

int lex_line;
int lex_tkstart;
char * srcfilename;
int verbose;		/* -v */
int raw;		/* -r */
int preparse;		/* -p */
int lbl;		/* -l  (line-by-line) */
int srcident;		/* -#  (emit #line directives in output) */
int maxsrcfiles = 255;	/* -i  (max distinct source files) */
int maxsrcnest = 10;	/* -j n (max include nesting) */
char * cont_str;	/* -c xxx */
int cont_len;		/*  length of above, set by init */
char * macro_str; 	/* -m xxx */
int macro_len;		/* length of -m string */
int cmntstr_len[MAX_COMMENT_STRS]; /* Length of -c strings */

int nincdirs;
char * incdirs[MAX_INCDIRS];		/* -I dir */

int ncmntstrs;
char * cmntstrs[MAX_COMMENT_STRS];	/* -c str */


/* Macro parsing states */
typedef enum
{
  mp_none = 0,
  mp_clspar,
  mp_tok_any,
  mp_tok_nospc,
  mp_opnpar,
  mp_macname,
  mp_dot
} mparse;

static mparse mp;


/* tokens returned from yylook, which recognises
   regular expressions in the input. */
typedef enum
{
  YYL_ERROR = -1,
  YYL_EOF = 0,
  YYL_PUNCT,	/* Punctuation represented by single char */
  YYL_IDENT,	/* Variable identifier */
  YYL_ILLEGAL,	/* Unrecognised char */
  YYL_HEX,	/* Hex number 0x.... */
  YYL_DEC,	/* Decimal number */
  YYL_FLOAT,	/* Floating point number */
  YYL_STRING,	/* String constant, including free text */
  YYL_CPPHASH,	/* Pre-processor hash */
  YYL_IF,
  YYL_ELSE,
  YYL_ELSEIF,
  YYL_ENDIF,
  YYL_FOR,
  YYL_FORALL,
  YYL_NEXT,
  YYL_KEEP,
  YYL_IFNONE,
  YYL_SORT,
  YYL_SWITCH,
  YYL_CASE,
  YYL_ENDSWITCH,
  YYL_METHOD,
  YYL_RETURN,
  YYL_THIS,
  YYL_OUTFILE,	/* ... keywords */
  YYL_GLOBAL,	/* :: */
  YYL_GE,	/* >= */
  YYL_LE,	/* <= */
  YYL_NE,	/* != */
  YYL_STRICTMATCHES,	/* ~= */
  YYL_SHL,
  YYL_SHR,
  YYL_LOGAND,		/* && */
  YYL_LOGOR,		/* || */
  YYL_EQ,		/* == */
  YYL_MATCHES,		/* ~ */
  YYL_SEMI,		/* ; */
  YYL_OPNPAR,		/* ( */
  YYL_CLSPAR,		/* ) */
  YYL_DBLSEMI,		/* ;; */
  YYL_REFERS,		/* <- */
  YYL_OPENIX,	/* [ OR (| */
  YYL_CLOSEIX,  /* ] OR |) */
  YYL_END_OF_TABLE
} looktkn;

static looktkn yylook(void);  /* forward */
static looktkn prevtk = YYL_ERROR;
static int ins_token;
static char quote;  /* If string, contains (closing) quote char */


typedef struct uniqfile_t
{
  char * name;		/* Canonical file name */
  struct stat st;	/* Stats: st_ino and st_dev uniquely identify the
  			   file. */
  int    numinc;	/* Number of times included */
} uniqfile;

static int nuniq;	    /* Number of entries in following */
static uniqfile * srcfiles; /* Array of maxsrcfiles entries */
  

typedef struct filestack_t
{
  FILE * f;
  int filenum;		/* Entry number in uniqfile list */
  int line;		/* Line number was up to */
  int morestring;	/* Free string state:
  			   0: no significance
                           1: Restart in free string mode
                           2: Send a semicolon token */
  int startup;		/* True only before 1st char read, or after yywrap(). */
} filestack;

static int fsp;		    /* File stack pointer.  When 1, we are
                               scanning the outermost source file.
                               fstack[fsp-1] is the current source file. */
static filestack * fstack;  /* Array of maxsrcnest entries */
static filestack * top;     /* Curent top of filestack */

static FILE * findopen(char * fn, int angleBkt, char ** whichdir)
{
  /* Search for file named fn.  If found, open it and return the file
     handle.  Search algorithm is:
     1. If fn starts with '/', then don't search (fn is absolute).
     2. If not angleBkt, attempt to locate fn in the current directory.
     3. Locate fn in the directory(s) specified in the incdirs (-I)
        array.
     4. If not found in any of the above, return null.
     5. If found in a -I directory (not cwd), set *whichdir to the
        directory filename.  Otherwise, don't touch *whichdir.
     6. Open file and return handle.  If error, print message and
        return null.
     Note that these are not the same semantics as the ANSI C preprocessor.
  */
  FILE * f;
  int i;
  char buf[513];
  
  if (fn[0] == '/')
  {
    f = fopen(fn, "r");
    if (!f && errno == ENOENT)
      goto notfound;
  }
  else
  {
    if (!angleBkt)
    {
      f = fopen(fn, "r");
      if (f)	/* found in cwd */
        return f;
    }
    for (i = 0; i < nincdirs; i++)
    {
      if (strlen(incdirs[i]) + strlen(fn) + 2 > sizeof(buf))
        /* Quietly ignore combinations which would exceed our buffer size */
        continue;
      *whichdir = incdirs[i];
      strcpy(buf, incdirs[i]);
      strcat(buf, "/");
      strcat(buf, fn);
      if (f = fopen(buf, "r"))
        break;
      else if (errno != ENOENT)
        break;
    }
    if (i == nincdirs)
    {
    notfound:
      /* failed to find file in any directory */
      fprintf(stderr, "%s:%d: file `%s' not found\n", 
        srcfilename, lex_line, fn);
      return 0;
    }
  }
  if (!f)
    /* There was an error other than 'not found'. */
    fprintf(stderr, "%s:%d: include file `%s': %s\n", 
        srcfilename, lex_line, fn, 
        errno < sys_nerr ? sys_errlist[errno] : "unexpected error");
  return f;
}

static void push_file(char * fn)
{
  /* fn points to file name to include, enclosed in either double quotes
     or <>.  File name does not require double quotes.  If quotes are
     omitted, semantics are same as if quoted.  Semantics of #include
     are specified above (in findopen()).
  */
  int angleBkt = 0;
  char trm = 0;
  char * p;
  int i;
  struct stat st;
  struct stat * stp;
  char * whichdir = "";	/* -I directory in which located, "" if cwd */
  
  if (fn[0] == '<')
  {
    if (fn[0] != '/') angleBkt = 1;
    fn++;
    trm = '>';
  }
  else if (fn[0] == '"')
  {
    fn++;
    trm = '"';
  }
  if (trm && (p = strchr(fn, trm)))
    *p = 0;
  while (isspace(*fn)) fn++;
  if (!*fn)
  {
    fprintf(stderr, "%s:%d: null #include ignored\n", srcfilename, lex_line);
    return;
  }
  while (isspace(fn[strlen(fn)-1]))
    fn[strlen(fn)-1] = 0;
  if (fsp >= maxsrcnest)
  {
    fprintf(stderr, "%s:%d: #include nesting greater than %d\n", 
      srcfilename, lex_line, maxsrcnest);
    return;
  }
  if (!(fstack[fsp].f = findopen(fn, angleBkt, &whichdir)))
    return;

  fstat(fileno(fstack[fsp].f), &st);
  for (i = 0; i < fsp; i++)
  {
    stp = &srcfiles[fstack[i].filenum].st;
    if (st.st_ino == stp->st_ino && st.st_dev == stp->st_dev)
    {
      fprintf(stderr, "%s:%d: recursive #include file `%s'\n", 
        srcfilename, lex_line, fn);
      fclose(fstack[fsp].f);
      return;
    }
  }
  for (i = 0; i < nuniq; i++)
  {
    stp = &srcfiles[i].st;
    if (st.st_ino == stp->st_ino && st.st_dev == stp->st_dev)
      break;
  }
  if (i == nuniq)
  {
    if (nuniq >= maxsrcfiles)
    {
      fprintf(stderr, "%s:%d: more than %d distinct #include files\n", 
        srcfilename, lex_line, maxsrcfiles);
      fclose(fstack[fsp].f);
      return;
    }
    memcpy(&srcfiles[i].st, &st, sizeof(struct stat));
    /* Construct 'canonical name' of this file. */
    if (*whichdir)
    {
      /* File located in one of the -I directories */
      srcfiles[i].name = (char *)malloc(strlen(whichdir)+1+strlen(fn)+1);
      strcpy(srcfiles[i].name, whichdir);
      strcat(srcfiles[i].name, "/");
      strcat(srcfiles[i].name, fn);
    }
    else
    {
      srcfiles[i].name = (char *)malloc(strlen(fn)+1);
      strcpy(srcfiles[i].name, fn);
    }
    nuniq++;
  }
  srcfiles[i].numinc++;  /* Count number of times included */
  fstack[fsp].startup = 1;
  fstack[fsp].filenum = i;
  fstack[fsp].morestring = fstack[fsp-1].morestring;
  fstack[fsp-1].line = lex_line;
  srcfilename = srcfiles[i].name;
  lex_line = 1;
  yyin = fstack[fsp].f;
  top = fstack + fsp;
  fsp++;
}

int yywrap(void)
{
  if (fsp == 1)
    return 1;
  fsp--;
  fclose(fstack[fsp].f);
  yyin = fstack[fsp-1].f;
  lex_line = fstack[fsp-1].line;
  srcfilename = srcfiles[fstack[fsp-1].filenum].name;
  top = fstack + fsp - 1;
  top->startup = 1;
  if (top->morestring)
  {
    bump_linect = 1;
    lex_line--;
  }
  return 0;
}

/*
 * We need to emulate the output of lex i.e. create a c function (yylex) 
 * which returns the next token in sequence.  Tokens are ints, which are
 * either single characters or one of the defines given in dgparse.y.  Note
 * that the yacc source explicitly includes this C source, which is
 * how the yacc token type defines are accessed.
 *
 * The structure yylval contains the semantic information for each token,
 * which should be set by yylex.
 *
 * The file pointer yyin is used to read the input.
 */

int yylex(void)
{
  int kilo;
  
  for(;;)
  {
  switch (yylook())
  {
case  YYL_ERROR: return 0;
case  YYL_PUNCT: return yytext[0];
case  YYL_IF: return IF;
case  YYL_ELSE: return ELSE;
case  YYL_ELSEIF: return ELSEIF;
case  YYL_ENDIF: return ENDIF;
case  YYL_FOR: return FOR;
case  YYL_FORALL: return FORALL;
case  YYL_NEXT: return NEXT;
case  YYL_KEEP: return KEEP;
case  YYL_IFNONE: return IFNONE;
case  YYL_SORT: return SORT;
case  YYL_SWITCH: return SWITCH;
case  YYL_CASE: return CASE;
case  YYL_ENDSWITCH: return ENDSWITCH;
case  YYL_METHOD: return METHOD;
case  YYL_RETURN: return RETURN;
case  YYL_THIS: return THIS;
case  YYL_OUTFILE: return OUTFILE;	/* ... keywords */
case  YYL_GLOBAL: return GLOBAL;	/* :: */
case  YYL_GE: return GE;		/* >= */
case  YYL_LE: return LE;		/* <= */
case  YYL_NE: return NE;		/* != */
case  YYL_STRICTMATCHES: return STRICTMATCHES;	/* ~= */
case  YYL_SHL: return SHL;		/* << */
case  YYL_SHR: return SHR;		/* >> */
case  YYL_LOGAND: return LOGAND;	/* && */
case  YYL_LOGOR: return LOGOR;		/* || */
case  YYL_EQ: return EQ;		/* == */
case  YYL_MATCHES: return MATCHES;	/* ~ */
case  YYL_SEMI: return ';';		/* ; */
case  YYL_OPNPAR: return '(';		/* ( */
case  YYL_CLSPAR: return ')';		/* ) */
case  YYL_DBLSEMI: return DBLSEMI;	/* ;; */
case  YYL_REFERS: return REFERS;	/* <- */
case  YYL_OPENIX: return '[';		/* [ OR (| */
case  YYL_CLOSEIX: return ']';  	/* ] OR |) */

case YYL_EOF:
		if (yywrap()) return(0);
		break;
case YYL_HEX:
		 if (toupper(yytext[yyleng-1]) == 'K')
		   kilo = 1, yytext[--yyleng] = 0;
		 else kilo = 0;
		 sscanf(yytext+2,"%x",&yylval.lex_int);
		 if (kilo)
		   yylval.lex_int <<= 10;
		 DEBUGPRINT(("integer `%d'\n", yylval.lex_int));
		 return INTEGER;
		
case YYL_DEC:
		 if (toupper(yytext[yyleng-1]) == 'K')
		   kilo = 1, yytext[--yyleng] = 0;
		 else kilo = 0;
		 yylval.lex_int = atoi(yytext);
		 if (kilo)
		   yylval.lex_int <<= 10;
		 DEBUGPRINT(("integer `%d'\n", yylval.lex_int));
		 return INTEGER;

case YYL_FLOAT:
/*
{D}+"."{D}*({E})?	|   e.g. 54.0E30 or 6.890 but not 7.E4
{D}*"."{D}+({E})?	|        .44E-08
{D}+{E}                          678E+8
*/
		 yylval.lex_float = atof(yytext);
		 DEBUGPRINT(("float `%lf'\n", yylval.lex_float));
		 return FLOATING;
		
case YYL_STRING:
		 yylval.lex_str = yylstr;
                 return STRING;

case YYL_CPPHASH:
		{		/* Hash created by C preprocessor CPP */
		  char c;
		  char str[1024];
		  char linestr[1024];
		  char * strp = str;
		  int i = 0;
		
		  while ((c=input()) && c != YYNEWLINE && 
		         strp - str < sizeof(str)-1)
		    *(strp++) = c;/* Read line info */
		  while (c != YYNEWLINE && (c=input()));
		  *strp = '\0';
		  DEBUGPRINT(("hash directive `%s'\n", str));
		  strp = str;
		  while (isspace(*strp))
		    strp++;   /* skip whitespace */
		  if (!strncmp(strp, "pragma", 6) && isspace(strp[6]))
		  {
		    /* Handle pragmas:
		       #pragma idem
		         - idempotent include file.  If inclusion count
		           for this file is >1, close and resume the
		           includer.
		    */
		    strp += 7;
		    while (isspace(*strp)) strp++;
		    if (!strncmp(strp, "idem", 4))
		    {
		      if (srcfiles[fstack[fsp-1].filenum].numinc > 1)
		        if (yywrap()) return(0);
		    }
		    goto end_cpphash;
		  }
		  if (!strncmp(strp, "include", 7) && isspace(strp[7]))
		  {
		    /* Include file */
		    strp += 8;
		    while (isspace(*strp)) strp++;
		    push_file(strp);
		    goto end_cpphash;
		  }
		  if (!strncmp(strp, "line", 4) && isspace(strp[4]))
		    strp += 5;
		  while (isspace(*strp) || isdigit(*strp))
		  {
		    linestr[i] = *strp;
		    i++;
		    strp++;
		  }
		  linestr[i] = '\0';
                  sscanf(linestr, " %d", &lex_line);
		  strcpy(srcfilename, strp);

		  if (verbose)
		    printf("Reading: line = %d, file = %s \n",
		            lex_line, srcfilename);
		end_cpphash:
		  if (c==YYNEWLINE)
		    unput(YYNEWLINE);
		}
		break;

case YYL_IDENT:
        	{
		  int i;
		  char uc = 0;
 
		  yylval.lex_str = (char *)malloc(yyleng+1);
		  strcpy(yylval.lex_str, yytext);
		  DEBUGPRINT(("identifier `%s'\n", yylval.lex_str));
		  return VAR;
		};

case YYL_ILLEGAL:
		fprintf(stderr,
		  "%s:%d: illegal char '%c' (%d = 0x%02X) ignored\n",
		  srcfilename, lex_line,
		  yytext[0], yytext[0], yytext[0]);
		break;
		
  } /* switch */
  
  } /* for ever */
  
  return 0;
}


static int parnest;	/* Parenthesis nesting when parsing macro parms */

/* This function recognises regular expressions in the input.  On entry,
   the next char is read from input.  On return, a character which is not
   whitespace but is not part of the previous token must be unput(). */
static looktkn yylook(void)
{
  unsigned char c, prev;
  unsigned short c2;
#define IS_ALPHA 256
#define IS_DIGIT 257
  char * p;
  looktkn tk = YYL_EOF;
  static looktkn nexttk;
  int se;
  int saveline;
  int newline = top->startup;
  mparse mpnext = mp;

#undef NEXT
#define NEXT (*p++ = c = input())
#define NEXTN (c = input())
#define UNDO(x) {p--; unput(x)}
#define UNDON(x) unput(x)

  p = yytext;
  top->startup = 0;
  
  if (top->morestring == 2)
  {
    top->morestring--;
    bump_linect = 1;
    tk = YYL_SEMI;
    goto bye;
  }
  else if (top->morestring)
  {
    top->morestring--;
    if (bump_linect)
    {
      bump_linect = 0;
      lex_line++;
      if (lbl)
      {
        NEXTN;
        if (c == '#')
        {
          top->morestring++;
          return YYL_CPPHASH;
        }
        UNDON(c);
      }
    }
    c = CLSBRACE;
    goto more_string;
  }
  
  if (mp == mp_dot)
  {
    tk = YYL_PUNCT;
    *p++ = '.';
    mpnext = mp_macname;
    goto bye;
  }
  
  if (ins_token)
  {
    ins_token = 0;
    tk = nexttk;
    goto bye_string;
  }
  
  NEXTN;

  if (isspace(c) && mp == mp_tok_nospc)
  {
    mpnext = mp_none;
    top->morestring = 1;
    UNDON(c);
    tk = YYL_CLSPAR;
    goto bye;
  }

  /* Skip whitespace and comments */
  for (;;)
  {
    while (isspace(c))
    {
      if (c == YYNEWLINE)
        newline = 1;
      NEXTN;
    }
    
    if (c == '/')
    {
      newline = 0;
      NEXTN;
      if (c == '/')
      {
        while (c && c != YYNEWLINE && c != CLSBRACE)
          NEXTN;
        if (c == CLSBRACE)  /* Close brace terminates comment and starts free
                               text string, even in // comment. */
          break;
      }
      else if (c == '*')
      {
        for (;;)
        {
          while (NEXTN)
            if (c == '*') break;
	  if (NEXTN == '/') break;
	}
	if (c)
	  NEXTN;
      }
      else
      {
        UNDON(c);
        c = '/';
        break;
      }
    }
    else
      break;
  }
  *p++ = c;

  if (isalpha(c))
    c2 = IS_ALPHA;
  else if (isdigit(c))
    c2 = IS_DIGIT;
  else
    c2 = c;
    
  lex_tkstart = lex_line;
  
  if (!c)
    goto check_insert_semi;
    
  switch(c2)
  {
    case IS_DIGIT:
      tk = YYL_DEC;
      if (c == '0')
      {
        NEXT;
        if (c == 'X' || c == 'x')
        {
          NEXT;
          if (!isxdigit(c))
          { UNDO(c); UNDO(*(p-1)); break; }
          while (isxdigit(c) && p - yytext < 10)
          {
            NEXT;
          }
          tk = YYL_HEX;
          goto check_suffix;
        }
        UNDO(c); 
        c = '0';
      }
      if (c == '0')
      {
        /* Don't store excess leading zeros */
        while (c == '0') NEXTN;
        UNDO(c);
        c = '0';
      }
      while (isdigit(c) && p - yytext < sizeof(yytext)-8) NEXT;
      if (isdigit(c))
      {
        /* Skip excess digits in integer or part to left of dec. point */
        while (isdigit(c)) NEXTN;
        UNDON(c);
        NEXT;
      }
      if (c == '.' || c == 'E' || c == 'e')
      {
        tk = YYL_FLOAT;
        if (c != '.')
        {
          /* Just integer plus exponent e.g. 123E-45 */
          se = 0;
          NEXT;
          if (!isdigit(c) && c != '+' && c != '-')
          {
            /* No exponent, back up to just integer part */
            UNDO(c); UNDO(*p);
            tk = YYL_DEC; break;
          }
          if (c == '+' || c == '-') se = 1, NEXT;
	  if (!isdigit(c))
	  { 
	    /* 'E' and sign, but no digits.  Back up to integer */
	    UNDO(c); UNDO(*p); 
	    if (se) UNDO(*p); 
	    break; 
	  }
	  else
	  {
            while (isdigit(c) && p - yytext < sizeof(yytext)-1) NEXT;
            while (isdigit(c)) NEXTN;  /* Ignore excessive exponent digits */
          }
        }
        else
        {
	  NEXT;
	  /* 6.E2 (e.g.) is ambiguous since it could be a float (600) or
	     a call to method E2 on the integer 6.  We resolve this by
	     not allowing floats in this form.  There must be at least
	     one digit after the decimal point. */
	  if (!isdigit(c))
	  { UNDO(c); UNDO('.'); tk = YYL_DEC; break; }
fract:
          se = 0;
	  while (isdigit(c) && p - yytext < sizeof(yytext)-7) NEXT;
	  if (isdigit(c))
	  {
	    /* Skip over excess fraction digits */
	    while (isdigit(c)) NEXTN;
	    UNDON(c);
	    NEXT;
	  }
	  if (c == 'E' || c == 'e')
	  {
	    /* Handle exponent in same manner as above */
	    NEXT;
	    if (!isdigit(c) && c != '+' && c != '-')
	    { UNDO(c); UNDO(*p); break; }
	    if (c == '+' || c == '-') se = 1, NEXT;
	    if (!isdigit(c))
	    { UNDO(c); UNDO(*p); if (se) UNDO(*p); break; }
	    else
	    {
	      while (isdigit(c) && p - yytext < sizeof(yytext)-1) NEXT;
	      while (isdigit(c)) NEXTN;
	    }
	  }
        }
      }
    check_suffix:
      if (c != 'k' && c != 'K')
        UNDO(c);
      break;
    case CLSBRACE:
    case '"':
    case '\'':
    case '`':
    more_string:
      tk = YYL_STRING;
      /* Open string quotation.  A quote of the same type
         within the string must be escaped with a backslash. */
      if (c == CLSBRACE)
        quote = OPNBRACE;
      else
        quote = c;
      p = yytext;
      newline = 0;

      for(;;)
      {
        NEXTN;
        if (!c || c == quote)
          break;
        if (c == YYNEWLINE && lbl && quote == OPNBRACE)
        {
          top->morestring = 2;
          /* Back off lex_line by 1 to get correct line number for
             the string pre-CR */
          lex_line--;
          c = quote;
          break;
        }
        if (c == '\\')
        {
          NEXTN;
          if (c != quote && c != '\\')
          {
            if (quote == '\'')
            {
              switch(c)
              {
              case 'a': c = '\a'; break;
              case 'b': c = '\b'; break;
              case 'f': c = '\f'; break;
              case 'n': c = '\n'; break;
              case 'r': c = '\r'; break;
              case 't': c = '\t'; break;
              case 'v': c = '\v'; break;
              case '0': c = CTRL_0; break;
              case '1': c = CTRL_1; break;
              case '2': c = CTRL_2; break;
              case '3': c = CTRL_3; break;
              case '4': c = CTRL_4; break;
              case '5': c = CTRL_5; break;
              case '6': c = CTRL_6; break;
              case '7': c = CTRL_7; break;
              case '8': c = CTRL_8; break;
              case '9': c = CTRL_9; break;
              default:
                UNDON(c);
                c = '\\';
              }
            }
            else
            {
              UNDON(c);
              c = '\\';
            }
          }
        }
        if (p - yytext < sizeof(yytext)-1)
        {
          *p++ = c;
              
          if (lbl && 		/* line-by-line mode only -l */
              quote == OPNBRACE && /* in free string */
              mp == mp_none &&	/* in initial macro parsing state */
              macro_len &&	/* we are looking for macros -m */
              p-yytext >= macro_len && /* Sufficient chars to recognise */
              !memcmp(p-macro_len, macro_str, macro_len)) /* is macro recog. */
          {
            /* May be macro expansion, but first check to see that string
               is not already commented according to the -c strings. */
            int i;
            int notcomment = 1;
            *p = 0;	/* Null term for strstr */
            for (i = 0; i < ncmntstrs; i++)
              if (strstr(yytext, cmntstrs[i]))
              {
                notcomment = 0;
                break;
              }
            if (notcomment)
            {
              p -= macro_len;
              while (p > yytext && isspace(p[-1])) p--;
              mpnext = mp_dot;
              c = quote;
              break;
            }
          }
        }
        else
        {
          if (p - yytext == sizeof(yytext)-1)
          {
            *p = c;
            if (fgetpos(yyin, &yynsc) < 0)
            {
              fprintf(stderr,
                 "%s:%d: Input not seekable; string length %d\n",
                 srcfilename, lex_line, yyleng);
              return YYL_ERROR;
            }
            saveline = lex_line;
          }
          p++;  /* Implicitly keep counting length */
        }
      }
      
      yyleng = p - yytext;
      yylstr = (char *)malloc(yyleng + 1);
      if (!yylstr)
      {
        fprintf(stderr, "%s:%d: Out of memory; string length %d\n",
           srcfilename, lex_line, yyleng);
        return YYL_ERROR;
      }
      if (yyleng < sizeof(yytext))
      {
        memcpy(yylstr, yytext, yyleng);
        yylstr[yyleng] = 0;
        if (c != quote)
          UNDON(c);
        if (newline)
          UNDON(YYNEWLINE);
      }
      else
      {
        memcpy(yylstr, yytext, sizeof(yytext));
        if (fsetpos(yyin, &yynsc) < 0)
        {
          fprintf(stderr,
                 "%s:%d: Input not seekable; string length %d\n",
                 srcfilename, lex_line, yyleng);
          return YYL_ERROR;
        }
        lex_line = saveline;
        /* Re-read rest of string */
        p = yylstr + sizeof(yytext);

        for(;;)
        {
          NEXTN;
          *p = 0;
          if (!c || c == quote)
            break;
          if (c == YYNEWLINE && lbl && quote == OPNBRACE)
          {
            c = quote;
            break;
          }
          if (c == '\\')
          {
            NEXTN;
            if (c != quote && c != '\\')
            {
	      if (quote == '\'')
	      {
                switch(c)
                {
                case 'a': c = '\a'; break;
                case 'b': c = '\b'; break;
                case 'f': c = '\f'; break;
                case 'n': c = '\n'; break;
                case 'r': c = '\r'; break;
                case 't': c = '\t'; break;
                case 'v': c = '\v'; break;
                case '0': c = CTRL_0; break;
                case '1': c = CTRL_1; break;
                case '2': c = CTRL_2; break;
                case '3': c = CTRL_3; break;
                case '4': c = CTRL_4; break;
                case '5': c = CTRL_5; break;
                case '6': c = CTRL_6; break;
                case '7': c = CTRL_7; break;
                case '8': c = CTRL_8; break;
                case '9': c = CTRL_9; break;
                default:
                  UNDON(c);
                  c = '\\';
                }
              }
              else
              {
                UNDON(c);
                c = '\\';
              }
            }
          }
          *p++ = c;
        }

        if (c != quote)
          UNDON(c);
        if (newline)
          UNDON(YYNEWLINE);
      }
      goto bye_string;
    case '#':
      if (newline)
      { tk = YYL_CPPHASH; break; }
      goto natl;
    case '.':
      NEXT;
      if (isdigit(c))
      { tk = YYL_FLOAT; goto fract; }
      UNDO(c);
      tk = YYL_PUNCT;
      break;
    case '(': 
      NEXT;
      if (c == '|')
      { tk = YYL_OPENIX; break; }
      UNDO(c);
      tk = YYL_OPNPAR;
      if (mp == mp_opnpar)
        mpnext = mp_tok_any, parnest = 1;
      else if (mp == mp_tok_any)
        parnest++;
      break;
    case '|': 
      NEXT;
      if (c == ')')
      { tk = YYL_CLOSEIX; break; }
      else if (c == '|')
      { tk = YYL_LOGOR; break; }
      UNDO(c);
      tk = YYL_PUNCT;
      break;
    case '[': tk = YYL_OPENIX; break;
    case ']': tk = YYL_CLOSEIX; break;
    case ':':
      NEXT;
      if (c == ':')
        tk = YYL_GLOBAL;
      else
        tk = YYL_ILLEGAL;
      break;
    case '>':
      NEXT;
      if (c == '=')
        tk = YYL_GE;
      else if (c == '>')
        tk = YYL_SHR;
      else
      { UNDO(c); tk = YYL_PUNCT; }
      break;
    case '<':
      NEXT;
      if (c == '=')
        tk = YYL_LE;
      else if (c == '<')
        tk = YYL_SHL;
      else if (c == '-')
        tk = YYL_REFERS;
      else
      { UNDO(c); tk = YYL_PUNCT; }
      break;
    case '=':
      NEXT;
      if (c == '=')
        tk = YYL_EQ;
      else
      { UNDO(c); tk = YYL_PUNCT; }
      break;
    case '!':
      NEXT;
      if (c == '=')
        tk = YYL_NE;
      else
      { UNDO(c); tk = YYL_PUNCT; }
      break;
    case '~':
      NEXT;
      if (c == '=')
        tk = YYL_STRICTMATCHES;
      else
      { UNDO(c); tk = YYL_MATCHES; }
      break;
    case '&':
      NEXT;
      if (c == '&')
        tk = YYL_LOGAND;
      else
      { UNDO(c); tk = YYL_PUNCT; }
      break;
    case ';':
      NEXT;
      if (c == ';')
        tk = YYL_DBLSEMI;
      else
      { UNDO(c); tk = YYL_SEMI; }
      break;
    case '+':
    case '/':
    case '-':
    case '*':
    case '%':
    case '^':
    case ',':
      tk = YYL_PUNCT; 
      break;
    case ')': 
      tk = YYL_CLSPAR;
      if (mp == mp_tok_any)
      {
        parnest--;
        if (!parnest)
        {
          mpnext = mp_none;
          top->morestring = 1;
        }
      }
      break;
    case IS_ALPHA:
    case '_':
    case '@':
    case '$':
natl:
      while(isalpha(c) ||
            isdigit(c) ||
            c == '@' || c == '#' || c == '$' || c == '_')
        if (p - yytext < sizeof(yytext)-1)
          NEXT;
        else
          NEXTN;
      
      UNDO(c);
      *p = 0;
      if (mp == mp_opnpar || mp == mp_tok_nospc)
      {
        /* When scanning unparenthesised macro parms, turn identifiers
           into strings. */
        tk = YYL_STRING;
        yylstr = (char *)malloc(strlen(yytext)+1);
        strcpy(yylstr, yytext);
        goto bye;
      }
      tk = YYL_IDENT;
      if (mp == mp_macname)
        mpnext = mp_opnpar;
      /* Check for keywords */
      switch (yytext[0])
      {
        case 'i':
          if (!strcmp(yytext+1, "f"))
            tk = YYL_IF;
          else if (!strcmp(yytext+1, "fnone"))
            tk = YYL_IFNONE;
          break;
        case 'e':
          if (!strcmp(yytext+1, "lse"))
            tk = YYL_ELSE;
          else if (!strcmp(yytext+1, "lseif"))
            tk = YYL_ELSEIF;
          else if (!strcmp(yytext+1, "ndif"))
            tk = YYL_ENDIF;
          else if (!strcmp(yytext+1, "ndswitch"))
            tk = YYL_ENDSWITCH;
          break;
        case 'f':
          if (!strcmp(yytext+1, "or"))
            tk = YYL_FOR;
          else if (!strcmp(yytext+1, "orall"))
            tk = YYL_FORALL;
          break;
        case 'k':
          if (!strcmp(yytext+1, "eep"))
            tk = YYL_KEEP;
          break;
        case 'n':
          if (!strcmp(yytext+1, "ext"))
            tk = YYL_NEXT;
          break;
        case 's':
          if (!strcmp(yytext+1, "ort"))
            tk = YYL_SORT;
          else if (!strcmp(yytext+1, "witch"))
            tk = YYL_SWITCH;
          break;
        case 'c':
          if (!strcmp(yytext+1, "ase"))
            tk = YYL_CASE;
          break;
        case 'm':
          if (!strcmp(yytext+1, "ethod"))
            tk = YYL_METHOD;
          break;
        case 'r':
          if (!strcmp(yytext+1, "eturn"))
            tk = YYL_RETURN;
          break;
        case 't':
          if (!strcmp(yytext+1, "his"))
            tk = YYL_THIS;
          break;
        case 'o':
          if (!strcmp(yytext+1, "utfile"))
            tk = YYL_OUTFILE;
          break;
      }
      if (tk != YYL_IDENT && tk != YYL_THIS && tk != YYL_KEEP &&
          tk != YYL_SORT && tk != YYL_IFNONE)
      {
check_insert_semi:
        /* If previous token was not a keyword or punctuation of any sort,
           insert a semicolon. */
        if (prevtk == YYL_STRING || prevtk == YYL_HEX ||
            prevtk == YYL_DEC || prevtk == YYL_FLOAT)
        {
          nexttk = tk;
          tk = YYL_SEMI;
          ins_token = 1;
        }
      }
      break;
    default:
      tk = YYL_ILLEGAL;
  }

bye:
  *p = 0;
  yyleng = p - yytext;

bye_string:
  if (tk != YYL_CPPHASH)
    prevtk = tk;
  
  if (mp == mp_opnpar && (tk != YYL_OPNPAR))
  {
    mpnext = mp_tok_nospc;
    nexttk = tk;
    tk = YYL_OPNPAR;
    ins_token = 1;
  }
  mp = mpnext;
  return tk;
}


int init_lex(int free_string, int skip_hashbang)
{
  /* (Re)initialise scanner.  free_string is 1 if start in free text mode
     i.e. as if '}' was just seen.  Otherwise starts in code mode. 
     If skip_hashbang is 1, then if the first 2 chars are '#!' then the
     first line is discarded. */
  int i;
  
  yysptr = yysbuf;
  prevtk = YYL_ERROR;
  ins_token = 0;
  lex_line = 1;
  parnest = 0;
  if (cont_str)
    cont_len = strlen(cont_str);
  else
    cont_len = 0;
  if (macro_str)
    macro_len = strlen(macro_str);
  else
    macro_len = 0;
  for (i = 0; i < ncmntstrs; i++)
    cmntstr_len[i] = strlen(cmntstrs[i]);
  mp = mp_none;
  if (skip_hashbang)
  {
    char c;
    if ((c = input()) == '#')
      if ((c = input()) == '!')
        while (input() != YYNEWLINE);
      else
      {
        unput(c);
        unput('#');
      }
    else
      unput(c);
  }
  /*if (free_string)
    unput(CLSBRACE);*/	/* This causes startup in 'free text scan' mode */
  
  if (srcfiles)
    free(srcfiles);
  if (fstack)
    free(fstack);
  srcfiles = (uniqfile *)malloc(sizeof(uniqfile) * maxsrcfiles);
  fstack = (filestack *)malloc(sizeof(filestack) * maxsrcnest);
  if (!srcfiles || !fstack)
    return 1;
  nuniq = 1;
  srcfiles[0].name = srcfilename;
  fstat(fileno(yyin), &srcfiles[0].st);
  fsp = 1;
  fstack[0].f = yyin;
  fstack[0].filenum = 0;
  fstack[0].line = lex_line;
  fstack[0].morestring = free_string ? 1 : 0;
  fstack[0].startup = 1;
  if (free_string)
  {
    lex_line--;
    bump_linect = 1;
  }
  top = fstack;
  return 0;
}

