/*****************************************************************************
*   $Id: dgvar.c,v 1.1.1.1 2000/06/22 15:39:56 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    techno@spirit.net.au
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: dgvar.c,v $
* Revision 1.1.1.1  2000/06/22 15:39:56  tekno
* Opened source
*
*
*****************************************************************************/
/* dgvar.c - variables, methods and storage management. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <unistd.h>	/* For dup2, pipe, fork */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <regex.h>
#include "dglex.h"
#include "dgvar.h"
#include "dgdb.h"
#include "y.tab.h"

extern expr_result eval_expr(int * ei, char ** es, tree t);

#define DG_METHOD(tag, type, ret, funcret, func, funcsig, nparms, name, desc)\
  funcret func funcsig;
#include "dgmethods.h"
#undef DG_METHOD

/* Method table for basic (int, string) methods */
static meth_tab_ent meth_tab[] =
{
#define DG_METHOD(tag, type, ret, funcret, func, funcsig, nparms, name, desc)\
  {ret, type, (int (*)(int, list))&func, nparms, name, desc },
#include "dgmethods.h"
#undef DG_METHOD
  {0, }
};

#define NUM_METHODS (sizeof(meth_tab)/sizeof(meth_tab[0])-1)

method translate_methodname(char * m)
{
  int i;
  
  for (i = 0; i < NUM_METHODS; i++)
    if (!strcmp(m, meth_tab[i].name))
      return (method)i;
  return M_CUSTOM;
}

int int_parm(list p, int parmno, int dflt)
{
  expr_result er;
  int ei;
  char * es;
  while (p && parmno)
    p = p->next, parmno--;
  if (!p)
    return dflt;
  er = eval_expr(&ei, &es, p->val);
  if (er == E_INT)
    return ei;
  return dflt;
}

char * str_parm(list p, int parmno, char * dflt)
{
  expr_result er;
  int ei;
  char * es;
  while (p && parmno)
    p = p->next, parmno--;
  if (!p)
    return dflt;
  er = eval_expr(&ei, &es, p->val);
  if (er == E_STRING)
    return es;
  return dflt;
}

var_node * var_parm(list p, int parmno, var_node * dflt)
{
  while (p && parmno)
    p = p->next, parmno--;
  if (!p)
    return dflt;
  if (p->val->type != T_VAR)
    return dflt;
  return (&p->val->u.var);
}

int m_length(char * s)
{
  return strlen(s);
}

char * m_match(var * v, list p)
{
  int x = int_parm(p, 0, 0);
  char * s;
  if (!(v->flags & VAR_SLICE))
    return "";
  if (x < 0 || x >= v->intval)
    return "";
  x = x * 2 + 1;
  s = tsmalloc(v->si[x].len + 1);
  strncpy(s, v->si[x].begin, v->si[x].len);
  s[v->si[x].len] = 0;
  return s;
}

char * m_not_match(var * v, list p)
{
  int x = int_parm(p, 0, 0);
  char * s;
  if (!(v->flags & VAR_SLICE))
    return "";
  if (x < 0 || x > v->intval)
    return "";
  x *= 2;
  s = tsmalloc(v->si[x].len + 1);
  strncpy(s, v->si[x].begin, v->si[x].len);
  s[v->si[x].len] = 0;
  return s;
}

char * m_unslice(var * v, list p)
{
  int x = int_parm(p, 0, 0);
  int i, l, o;
  char * s;
  if (!(v->flags & VAR_SLICE))
    return "";
  if (x <= 0 || x > v->intval * 2 + 1)
    x = v->intval * 2 + 1;
  for (i = 0, l = 0; i < x; i++)
    l += v->si[i].len;
  s = tsmalloc(l + 1);
  for (i = 0, o = 0; i < x; i++)
  {
    memcpy(s+o, v->si[i].begin, v->si[i].len);
    o += v->si[i].len;
  }
  s[l] = 0;
  return s;
}

#define DD(x) fprintf(stderr, x);

int m_slice(char * s, list p)
{
  /* 2 parms: regexp pattern, variable to place results (and turn into
     a so-called 'slice-var' */
  tree pn = p ? p->val : 0;
  var_node * vn = var_parm(p, 1, 0);
  int max_match = int_parm(p, 2, 0);
  var * v;
  regex_t reg;
  regex_t * preg = 0;
  char * patt;
  regmatch_t pm;
  int code;
  int nmatch = 0;
  long tmark, lasto;
  slice_info * si;
  
  if (!vn || !pn)
  {
    return 0;	/* insufficient parms */
  }
  v = existing_var(vn);
  if (v && v->flags & VAR_FORALL)
  {
    error(p->next->val, "Cannot use forall variable %s as slice var",
      v->name);
    return 0;   /* cannot convert forall to slice variable */
  }
  if (!v)
    v = create_var(vn);
  v->flags |= VAR_SLICE;
  
  if (pn->type == T_REGEXP)
    preg = (regex_t *)pn->u.rx;
  else
  {
    patt = str_parm(p, 0, 0);
    if (patt)
    {
      preg = &reg;
      if ((code = regcomp(preg, patt, REG_EXTENDED | REG_NEWLINE)))
        return 0; /*error_in_pattern(t, preg, code); */
      if (pn->type == T_STRING)
      {
        /* Constant string.  For speed, replace with precompiled rx node */
        pn->type = T_REGEXP;
        pn->u.rx = (struct re_pattern_buffer *)malloc(sizeof(regex_t));
        memcpy(pn->u.rx, preg, sizeof(regex_t));
        preg = (regex_t *)pn->u.rx;
      }
    }
    else
      return 0; /* rx not string parm */
  }
  
  if (v->flags & VAR_SLICE)
  {
    /* Recycling old slice var */
    if (v->strval && v->alloc) free(v->strval);
    if (v->si) free(v->si);
  }

  v->strval = (char *)malloc(v->alloc = strlen(s) + 1);
  strcpy(v->strval, s);
  
  tmark = mark_tempstack();
  lasto = 0;
  while (!regexec(preg, s + lasto, 1, &pm, 0))
  {
    nmatch++;
    si = (slice_info *)tsmalloc(sizeof(slice_info) * 2);
    si->begin = v->strval + lasto;
    si->len = pm.rm_so;
    si[1].begin = v->strval + lasto + pm.rm_so;
    si[1].len = pm.rm_eo - pm.rm_so;
    lasto += pm.rm_eo;
    if (!si[1].len)
      /* Matched zero chars, break out of potential infinite loop */
      break;
    if (nmatch == max_match)
      break;
  }
  
  si = (slice_info *)tsmalloc(sizeof(slice_info));
  si->begin = v->strval + lasto;
  si->len = strlen(si->begin);
  si -= nmatch * 2;	/* Point to first of constructed array */
  v->si = (slice_info *)malloc(lasto = sizeof(slice_info) * (nmatch*2 + 1));
  memcpy(v->si, si, lasto);
  v->nslice = nmatch*2 + 1;
  v->flags |= VAR_READONLY;
  
  pop_tempstack(tmark);
  
  v->intval = nmatch;
  if (preg == &reg)
    regfree(preg);
    
  return nmatch;
}

char * m_substr(char * s, list p)
{
  int p1 = int_parm(p, 0, 0);
  int p2 = int_parm(p, 1, -1);
  int l = strlen(s);
  char * r;
  
  if (p1 < 0)
    p1 = 0;
  if (p1 >= l)
  {
    r = tsmalloc(1);
    *r = 0;
  }
  else if (p2 < 0 || p1 + p2 > l)
  {
    r = tsmalloc(l - p1 + 1);
    strcpy(r, s + p1);
  }
  else
  {
    r = tsmalloc(p2 + 1);
    strncpy(r, s + p1, p2);
    r[p2] = 0;
  }
  return r;
}

static char * subs_page(char * s, int page, int line)
{
  /* Return header string, with substitutions.  %page% gets turned into
     page number in decimal -- only for strings less than 240 chars long. */
  char * p = strstr(s, "%page%");
  static char buf[256];
  if (p && strlen(s) < 240)
  {
    memcpy(buf, s, p-s);
    sprintf(buf + (p-s), "%d", page+1);
    strcat(buf, p + 6);
    return buf;
  }
  else
    return s;
}

char * m_page(char * s, list p)
{
  int lines = int_parm(p, 0, 60);	/* Number of lines per page */
  int hlines = int_parm(p, 1, 4);	/* Number of header lines */
  int flines = int_parm(p, 2, 0);	/* Number of footer lines */
  int max_group = int_parm(p, 3, 0);	/* Max line group */
  int l = strlen(s);
  char * r = s;
  int i;
  int wasflush = 0;
  int flush_group = 0;
  /* Internal state */
  static int linect;	/* User lines on current page (start at zero,
                           for first line after header) */
  static int pagect;	/* Page number (starting at zero) */
  static int ng;	/* Current buffered lines */
  static int group;	/* Whether in group */
#define MAX_GRP 60
  static char * gbuf[MAX_GRP]; /* Array of pointers to buffered lines */
  int hdrlen = 0;
  int ftrlen = 0;
#define MAX_PAGE_HDRS 10
  static char * hdrs[MAX_PAGE_HDRS];
  static char * ftrs[MAX_PAGE_HDRS];
  
  if (!l)
  {
    s = "\f-";
    l=2;	/* Flush call; equivalent to page break. */
    /* Reset state for next time */
    wasflush = 1;
  }
  /* Make sure some semblance of sanity */
  if (hlines < 0) hlines = 0;
  if (hlines > MAX_PAGE_HDRS) hlines = MAX_PAGE_HDRS;
  if (flines < 0) flines = 0;
  if (flines > MAX_PAGE_HDRS) flines = MAX_PAGE_HDRS;
  if (lines < hlines+flines+1)
    lines = hlines+flines+1;
  if (max_group < 1) max_group = 1;
  else if (max_group > lines-hlines-flines)
    max_group = lines-hlines-flines;
  if (max_group > MAX_GRP)
    max_group = MAX_GRP;
    
  if (s[0] == '\f')
  {
    /* Control code for us */
    int newpage = 0, set_hdr = -1, set_ftr = -1, reset_page = 0;
    s++;
    if (s[0] == '-') newpage = 1, s++;
    else if (s[0] == '0') newpage = 1 + (pagect & 1), s++;
    else if (s[0] == '1') newpage = 1 + !(pagect & 1), s++;
    if (s[0] == 'r') reset_page = 1, s++;
    if (s[0] == '(') group++, s++;
    if (s[0] == ')') 
    {
      group--;
      if (!group)
        flush_group = 1;
      else if (group < 0)
        group = 0;
      s++;
    }
    l = 0;
    if (newpage)
    {
      if (linect)
      {
        l = lines - flines - hlines - linect;
        for (i = 0; i < flines; i++)
          l += ftrs[i] ? strlen(subs_page(ftrs[i], pagect, 0)) : 1;
      }
      else
        l = lines;
      if (newpage > 1)
        l += lines;
      r = tsmalloc(l + 1);
      r[0] = 0;
      if (linect)
      {
        for (i = 0; i < lines - flines - hlines - linect; i++)
          strcat(r, "\n");
        for (i = 0; i < flines; i++)
          strcat(r, ftrs[i] ? subs_page(ftrs[i], pagect, 0) : "\n");
      }
      else
        for (i = 0; i < lines; i++)
          strcat(r, "\n");
      pagect++;
      if (newpage > 1)
      {
        pagect++;
        for (i = 0; i < lines; i++)
          strcat(r, "\n");
      }
      linect = 0;
    }
    else
      r = "";
    
    if (s[0] == 'h') set_hdr = s[1]-'0'-1, s += 2;
    else if (s[0] == 'f') set_ftr = s[1]-'0'-1, s += 2;
    if (set_hdr >= 0 && set_hdr < MAX_PAGE_HDRS)
    {
      if (hdrs[set_hdr]) free(hdrs[set_hdr]);
      hdrs[set_hdr] = (char *)malloc(strlen(s)+1);
      strcpy(hdrs[set_hdr], s);
    }
    if (set_ftr >= 0 && set_ftr < MAX_PAGE_HDRS)
    {
      if (ftrs[set_ftr]) free(ftrs[set_ftr]);
      ftrs[set_ftr] = (char *)malloc(strlen(s)+1);
      strcpy(ftrs[set_ftr], s);
    }
    if (reset_page)
      pagect = 0;
  }
  else if (!m_isctrl(s))
  {
    if (!linect)
      for (i = 0; i < hlines; i++)
        hdrlen += hdrs[i] ? strlen(subs_page(hdrs[i], pagect, 0)) : 1;
    if (linect == lines - hlines - flines - 1)
      for (i = 0; i < flines; i++)
        ftrlen += ftrs[i] ? strlen(subs_page(ftrs[i], pagect, i)) : 1;
    r = tsmalloc(hdrlen + l + ftrlen + 1);
    r[0] = 0;
    if (hdrlen)
      for (i = 0; i < hlines; i++)
        strcat(r, hdrs[i] ? subs_page(hdrs[i], pagect, 0) : "\n");
    strcat(r, s);
    if (ftrlen)
      for (i = 0; i < flines; i++)
        strcat(r, ftrs[i] ? subs_page(ftrs[i], pagect, 0) : "\n");
    linect++;
    if (linect == lines - hlines - flines)
      linect = 0, pagect++;
  }
  if (wasflush)
  {
    linect = 0;
    pagect = 0;
    for (i = 0; i < MAX_PAGE_HDRS; i++)
    {
      if (hdrs[i]) free(hdrs[i]), hdrs[i] = 0;
      if (ftrs[i]) free(ftrs[i]), ftrs[i] = 0;
    }
    group = 0;
    ng = 0;
  }
  return r;
}

static char * str_fit(char * s, int p1, int p2, char * p3)
{
  /* Fit string into fixed width field.  p1 is field length (default 7).
     Max allowed width is 1024.
     p2 is 0 for left justify, 3 for centre, 1 for right (dflt 0 = left).
     p3 is fill string (default ' ').
     If string is longer than field, it is truncated on the right.  Otherwise,
     it is justified into the field as specified, plus padding characters
     are added.  New string of specified length is allocated on tstack. */
  int l = strlen(s);
  int m;
  char * r;
  int i;
  
  if (p1 <= 0)
    return "";
  if (p1 > 1024)
    p1 = 1024;
  r = tsmalloc(p1+1);
  r[p1] = 0;
  if (l >= p1)
  {
    memcpy(r, s, p1);
    return r;
  }
  m = strlen(p3);
  if (!m)
    m = 1, p3 = " ";
  if (m == 1)
    memset(r, *p3, p1);
  else
    for (i = 0; i < p1; i += m)
      memcpy(r+i, p3, m < p1-i ? m : p1-i);
  if (p2 <= 0 || p2 == 2 || p2 > 3)
    memcpy(r, s, l);
  else if (p2 == 1)
    memcpy(r + p1 - l, s, l);
  else
    memcpy(r + ((p1 - l)>>1), s, l);
  return r;
}

static int next_int(char ** sp, int dflt)
{
  char * s = *sp;
  char * t;
  int i;
  
  while (isspace(*s)) s++;
  if (!*s)
    i = dflt;
  else
  {
    t = s;
    if (*s == '-' || *s == '+') s++;
    while (isdigit(*s)) s++;
    sscanf(t, "%d", &i);
  }
  *sp = s;
  return i;
}

#define MAX_TAB 256
#define JUST_LEFT	0x00
#define JUST_RIGHT	0x01
#define JUST_RIGHT_NOS	0x02
#define JUST_CENTRE	0x03
#define JUST_MASK	0x03
#define JUST_FORCE	0x04
#define JUST_NODELIM	0x08
#define JUST_TRUNCATE	0x10


static void add_tab(int * c, int * j, int * N, int col, int just, int width)
{
  int offs, grp, nc, mod;
  if (*N >= MAX_TAB)
    return;
  if (col < 0)
  {
    col = -col;
    if (col > *N) col = *N;
    offs = just;
    grp = *N - col;
    mod = 0;
    while (*N < MAX_TAB)
    {
      nc = c[grp+mod] + offs;
      while (nc < 1) nc += width;
      while (nc > width) nc -= width;
      c[*N] = nc;
      j[*N] = j[grp+mod];
      (*N)++;
      mod++;
      if (mod == col)
      {
        offs += just;
        mod = 0;
      }
    }
  }
  else
  {
    if (col > width)
      col = (col-1) % width + 1;
    c[*N] = col;
    j[*N] = just;
    (*N)++;
  }
}

static int col_space(int * c, int seg, int width)
{
  int w = c[seg+1] - c[seg];
  while (w <= 0) w += width + 1;
  return w;
}

char * m_tab(char * s, list p)
{
  int l = strlen(s);
  char * r = s;
  static int setup;	/* Set when tabs setup from parms */
  int col, just, i;
  static int width;
  static int c[MAX_TAB];
  static int j[MAX_TAB];
  static int N;
  
  if (!l)
  {
    setup = 0;
    return "";	/* Nothing buffered, propagate flush. */
  }
  if (!setup)
  {
  do_setup:
    width = int_parm(p, 0, 80);
    if (width < 1) width = 1;
    if (width > 1024) width = 1024;
    i = 1;
    N = 0;
    while ((col = int_parm(p, i++, 0)))
    {
      just = int_parm(p, i++, 0);
      add_tab(c, j, &N, col, just, width);
    }
    if (!N)
    {
      /* Force at least one tab stop (col 1, left) otherwise nothing comes out*/
      N = 1;
      c[0] = 1;
      j[0] = 0;
    }
    setup = 1;
    /*
    fprintf(stderr, "tab: width = %d\n", width);
    for (i = 0; i < N; i++)
      fprintf(stderr, "tab: %2d  c=%d j=%d\n", i, c[i], j[i]);
    */
  }
  if (s[0] == '\v')
  {
    /* tab control */
    if (s[1] != '\v')
      return s+1;	/* No tabulation */
    s += 2;
    if (*s == '\n') goto do_setup;
    width = next_int(&s, 80);
    if (width < 1) width = 1;
    if (width > 1024) width = 1024;
    N = 0;
    while ((col = next_int(&s, 0)))
    {
      just = next_int(&s, 0);
      add_tab(c, j, &N, col, just, width);
    }
    return "";
  }
  else if (!m_isctrl(s))
  {
    /* Line for us to tabulate.  Divide into segments delimited by \t
       chars.  Strip trailing space from each segment. */
    char * t;
    int seg;	/* Current segment no. */
    int nseg;	/* Actual number of segments (<= N) */
    int fl;	/* Total allocated length upper bound */
    char * st[MAX_TAB];
    int ll[MAX_TAB];
    char * u;
    int f;	/* Offset in return buffer of start of current line */
    int g;	/* Add to f to give offset of current field */
    int cp;	/* Previous col number */
    int cc;	/* This column number */
    int h;	/* Target start of next field */
    
    if (*s == '\n')
      return s;
    fl = width;
    for (seg = 0; seg < N; seg++)
    {
      t = s;
      /* Note: guaranteed terminating \n, but check for null anyway */
      while (*t && (seg == N-1 ? 1 : *t != '\t') && *t != '\n') t++;
      u = st[seg] = tsmalloc(t - s + 1);
      memcpy(u, s, l = t-s);
      while (l && isspace(u[l-1])) l--;
      u[l] = 0;
      ll[seg] = l;
      if (seg == N-1)
      {
        fl += l + 2;
        while (l--)
          if (u[l] == '\t')
            u[l] = ' ';
      }
      else
        if (l >= col_space(c, seg, width))
          fl += l + width;
        else
          fl += col_space(c, seg, width) + 2;
      fl += ll[seg]/width;
      if (*t == '\t')
        s = t+1;
      else
      {
        seg++;
        break;
      }
    }
    nseg = seg;
    /*
    fprintf(stderr, "tab: width=%d nseg=%d fl=%d\n", width, nseg, fl);
    for (i = 0; i < nseg; i++)
      fprintf(stderr, "tab: seg %d = `%s'\n", i, st[i]);
    */
    r = tsmalloc(fl+2);
    memset(r, ' ', fl);
    r[fl+1] = 0;
    f = g = cp = 0;
    for (seg = 0; seg < nseg; seg++)
    {
      cc = c[seg];
      if (cc <= cp)
      {
        /* Gone to new line */
        r[f+g] = '\n';
        f += g + 1;
        g = cc - 1;
      }
      else if (cc - 1 > g || j[seg] & JUST_FORCE)
        g = cc - 1;
      else if (g && !(j[seg] & JUST_NODELIM))
        g++;
      cp = cc;
      u = st[seg];
      if (seg == nseg-1 || c[seg+1] <= c[seg])
        /* Last field on line */
        h = width+2;
      else
      {
        if (j[seg] & JUST_TRUNCATE)
          h = c[seg+1];
        else
          h = 9999;
        switch(j[seg] & JUST_MASK)
        {
        case JUST_RIGHT:
          if (g + ll[seg] + 2 < c[seg+1])
            g = c[seg+1] - ll[seg] - 2;
          break;
        case JUST_RIGHT_NOS:
          if (g + ll[seg] + 1 < c[seg+1])
            g = c[seg+1] - ll[seg] - 1;
          break;
        case JUST_CENTRE:
          if (g + ll[seg] + 1 < c[seg+1])
            g = c[seg] + (c[seg+1]-c[seg])/2 - ll[seg]/2 - 1;
          break;
        default:
          break;
        }
      }
      while (*u)
      {
        if (g >= width)
        {
          h = 9999;  /* never get there */
          r[f+g] = '\n';
          f += g + 1;
          g = 0;
        }
        if (g >= h-1)
          break;
        r[f+g++] = *u++;
      }
    }
    r[f+g] = '\n';
    r[f+g+1] = 0;
    if (f+g > fl)
      fprintf(stderr, "tab: field overflow!\n");
  }
  return r;
}

char * m_fit(char * s, list p)
{
  /* Method interface to str_fit() */
  int p1 = int_parm(p, 0, 7);
  int p2 = int_parm(p, 1, 0);
  char * p3 = str_parm(p, 2, " ");
  return str_fit(s, p1, p2, p3);
}

char * m_contents(char * s, list p)
{
  /* Read file contents.  s contains file name, */
  int l = strlen(s);
  char * r;
  FILE * f;
  struct stat st;
  
  if (stat(s, &st))
      goto erret;
  f = fopen(s, "r");
  if (!f)
      goto erret;
  r = tsmalloc(st.st_size + 1);
  fread(r, 1, st.st_size, f);
  fclose(f);
  r[st.st_size] = 0;
  
  return r;
erret:
  r = tsmalloc(1);
  *r = 0;
  return r;
}

char * m_run(char * s, list p)
{
  /* Execute file named 's' and return resulting stdout
     as a string.  Optional string parm may be used to specify stdin
     to the command.  Default is zero-length stdin.  's' can contain
     parameters, delimited by spaces, which are passed as args to the
     program.  Args can also be delimited by single or double quotes
     if they contain spaces.  */
  char * in = str_parm(p, 0, "");
  char * out;
  int pd[2];	/* Pipe descriptors */
  pid_t ch;	/* Child PID */
  int st;	/* Child status */
  int nb;
  char * t, * u, ** a, quote;
  struct stat fs;
  
  if (pipe(pd))
  {
    perror("pipe()");
    return "";
  }
  ch = fork();
  if (ch == (pid_t)-1)
  {
    perror("fork()");
    return "";
  }
  if (!ch)
  {
    /* This is child */
    if (dup2(pd[1], fileno(stdout)) < 0)
    {
      perror("dup2(stdout)");
      exit(8);
    }
    if (dup2(pd[0], fileno(stdin)) < 0)
    {
      perror("dup2(stdin)");
      exit(8);
    }
    for (t = s; *t && !isspace(*t); t++);
    nb = 0;
    if (*t)
      *t++ = 0;
    for (u = t; *u; u++)
    {
      if (isspace(*u)) continue;
      nb++;
      if (*u == '\'' || *u == '"') quote = *u++;
      else quote = 0;
      while (*u && (quote ? *u != quote : !isspace(*u))) u++;
      if (*u) u++;
    }
    a = (char **)malloc(sizeof(char *) * (nb+2));
    a[0] = s;
    nb = 0;
    for (u = t; *u; u++)
    {
      if (isspace(*u)) continue;
      nb++;
      if (*u == '\'' || *u == '"') quote = *u++;
      else quote = 0;
      a[nb] = u;
      while (*u && (quote ? *u != quote : !isspace(*u))) u++;
      if (*u) *u++ = 0;
    }
    a[++nb] = 0;
    execvp(s, a);
    perror("execvp");
    exit(8);
  }
  else
  {
    write(pd[1], in, strlen(in));
    close(pd[1]);
    while (waitpid(ch, &st, 0) < 1)
      perror("waitpid()");
    fstat(pd[0], &fs);
    out = tsmalloc(fs.st_size+1);
    nb = read(pd[0], out, fs.st_size);
    if (nb < 0)
    {
      perror("read()");
      return "";
    }
    close(pd[0]);
    out[fs.st_size] = 0;
  }
  
  return out;
}

char * m_env(char * s, list p)
{
  /* Return the value of environment variable 's' as a string. */
  char * e = getenv(s);
  char * v;
  if (e)
  {
    v = tsmalloc(strlen(e)+1);
    strcpy(v, e);
    return v;
  }
  return "";
}

char * m_toupper(char * s)
{
  /* Return the string converted to uppercase. */
  char * v;
  char * p;
  p = v = tsmalloc(strlen(s)+1);
  while (*s)
    *p++ = toupper(*s++);
  *p = 0;
  return v;
}

char * m_tolower(char * s)
{
  /* Return the string converted to lowercase. */
  char * v;
  char * p;
  p = v = tsmalloc(strlen(s)+1);
  while (*s)
    *p++ = tolower(*s++);
  *p = 0;
  return v;
}

int m_int(char * s)
{
  int x = 0;
  sscanf(s, " %d", &x);
  return x;
}

int m_isint(char * s)
{
  while (isspace(*s)) s++;
  while (isdigit(*s)) s++;
  while (isspace(*s)) s++;
  if (*s)
    return 0;
  else
    return 1;
}

int m_hexint(char * s)
{
  int x = 0;
  if (s[0] == '0' && (s[1] == 'X' || s[1] == 'x')) s += 2;
  sscanf(s, " %x", &x);
  return x;
}

int m_ishexint(char * s)
{
  while (isspace(*s)) s++;
  if (s[0] == '0' && (s[1] == 'X' || s[1] == 'x')) s += 2;
  while (isxdigit(*s)) s++;
  while (isspace(*s)) s++;
  if (*s)
    return 0;
  else
    return 1;
}


char * m_inthex(int v, list p)
{
  char * s;
  char buf[20];
  int p1 = int_parm(p, 0, 0);	/* Precision - 0 = variable, else min # digits*/
  if (p1 > 0 && p1 < sizeof(buf))
    sprintf(buf, "%0*X", p1, v);
  else
    sprintf(buf, "%X", v);
  s = tsmalloc(strlen(buf)+1);
  strcpy(s, buf);
  return s;
}

int m_char(char * s)
{
  
  return (int)(unsigned char)s[0];
}

int m_isctrl(char * s)
{
  char c = s[0];
  switch (c)
  {
    case '\a': case '\b': case '\f': case '\v':
    case CTRL_0: case CTRL_1: case CTRL_2: case CTRL_3: case CTRL_4: 
    case CTRL_5: case CTRL_6: case CTRL_7: case CTRL_8: case CTRL_9:
      return 1;
  }
  return 0;
}

int m_is_first(var * v)
{
  return (int)(v->flags & VAR_FIRST);
}

int m_is_last(var * v)
{
  return (int)(v->flags & VAR_LAST);
}

expr_result user_method(int * ei, char ** es, method_info * mi, tree t)
{
  /* Call user-defined method.  mi->func points to T_FUNCTION node describing
     the method, parms, and return expression.  Actual parms which are
     not provided are given default values of int 0 or whatever the
     initialiser expression result.  Excess actual parms
     are ignored (not evaluated). */
  long tstack = mark_tempstack();
  expr_result er;
  char * s;	/* Result string */
  list formal = mi->func->u.mdef.parms;
  list actual = t->u.meth.parms;
  var * v;
  int fpi;	/* Formal parm int result */
  char * fps;	/* Formal parm string result */
  tree ttype = mi->func->u.mdef.thistype;	/* thistype */
  tree obj = t->u.meth.obj;	/* Object for which method called */
  var_node vn;	/* Temp node for creation of 'this' */
  int argno = 0;
  
  push_scope();
  while (actual && formal)
  {
    argno++;
    if (formal->val->type == T_REFERS)
    {
      /* Call by reference.  Formal parm is an alias of the actual parameter,
         which must be a variable. */
      if (actual->val->type != T_VAR)
      { er = error(t, "actual parameter #%d is not a variable for call by ref",
                   argno);
        goto bye; }
      if (!actual->val->u.var.var)
      /*{ er = error(t, 
          "actual parameter #%d is not initialised for call by ref", argno);
        goto bye; }*/
      {
        pseudo_pop_scope();
        actual->val->u.var.var = lookup_var(&actual->val->u.var);
        pseudo_push_scope();
        if (!actual->val->u.var.var)
        { er = error(t, "actual parameter #%d not defined for call by ref",
                 argno);
          goto bye; }
        
      }
      v = alias_var(actual->val->u.var.var, &formal->val->u.expr.left->u.var);
      if (!v)
      { er = error(t, "formal parameter duplicate (arg #%d)", argno);
        goto bye; }
    }
    else
    {
      v = create_var(&formal->val->u.expr.left->u.var);
      if (!v)
      { er = error(t, "formal parameter duplicate (arg #%d)", argno); 
        goto bye; }
      pseudo_pop_scope();
      /* Evaluate in caller's scope */
      er = eval_expr(&fpi, &fps, actual->val);
      pseudo_push_scope();
      if (er == E_ERROR) goto bye;
      if (er == E_INT)
      {
        v->type = E_INT;
        v->intval = fpi;
      }
      else
      {
        v->type = E_STRING;
        v->strval = fps;
      }
    }
    actual = actual->next;
    formal = formal->next;
  }
  while (formal)
  {
    argno++;
    v = create_var(&formal->val->u.expr.left->u.var);
    if (!v)
    { er = error(t, "formal parameter duplicate (arg #%d)", argno); goto bye; }
    if (formal->val->u.expr.right)
    {
      /* Initialiser expression provided.  Evaluate in current scope */
      er = eval_expr(&fpi, &fps, formal->val->u.expr.right);
      if (er == E_ERROR)
        goto bye;
      if (er == E_INT)
      {
        v->type = E_INT;
        v->intval = fpi;
      }
      else
      {
        v->type = E_STRING;
        v->strval = fps;
      }
    }
    else
    {
      /* Ultimate default of integer 0 */
      v->type = E_INT;
      v->intval = 0;
    }
    formal = formal->next;
  }
  
  if (ttype->type == T_VAR)
  {
    /* Method is for data type.  Check that object is a variable, and make
       'this' an alias of that variable. */
    if (obj->type != T_VAR)
    { er = error(t, "method %s requires a variable object", mi->name); 
      goto bye; }
    pseudo_pop_scope();
    v = lookup_var(&obj->u.var);
    pseudo_push_scope();
  }
  else
  {
    /* Method is for int or string.  Check matching object, and make a
       new variable called 'this' with a copy of the object */
    pseudo_pop_scope();
    er = eval_expr(&fpi, &fps, obj);
    pseudo_push_scope();
    if (er == E_ERROR) goto bye;
    vn.varname = "this";
    vn.var = 0;
    vn.global = 0;
    v = create_var(&vn);
    if (!v)
    { er = error(t, "'this' already exists as global in method %s", mi->name);
      goto bye; }
    if (er == E_INT)
    {
      v->type = E_INT;
      v->intval = fpi;
    }
    else
    {
      v->type = E_STRING;
      v->strval = fps;
    }
  }
  set_thisvar(v);
  execute(mi->func->u.mdef.stmts);
  er = eval_expr(ei, &s, mi->func->u.mdef.ret);

bye:
  pop_tempstack(tstack);
  if (er == E_STRING)
  {
    int len;
    *es = tsmalloc(len = strlen(s)+1);
    memmove(*es, s, len);
  }
  pop_scope();
  return er;
}

expr_result custom_method(int * ei, char ** es, var * v, char * mn, tree t)
{
  meth_tab_ent * mte = v->cht ? (meth_tab_ent *)lookup_hash(v->cht, mn) : 0;
  
  if (!mte)
  {
    method_info * mi = get_usermethod(mn);
    if (mi)
      return user_method(ei, es, mi, t);
    return error(t, "custom method %s not found", mn);
  }
  v->mte = mte;
  if (mte->ret == E_INT)
  {
    *ei = mte->func((int)v, t->u.meth.parms);
    return E_INT;
  }
  else
  {
    *es = (char *)mte->func((int)v, t->u.meth.parms);
    return E_STRING;
  }
}

expr_result invoke_method(int * ei, char ** es, tree t)
{
  /* Let's assume sizeof(int) == sizeof(char *)... */
  meth_tab_ent * mte = meth_tab + t->u.meth.name;
  int thisi;
  char * thiss;
  var_node * vn;
  var * v;
  expr_result er;
  method_info * mi;
  
  if (t->u.meth.name == M_CUSTOM)
  {
    if (t->u.meth.obj->type != T_VAR)
    {
      mi = get_usermethod(t->u.meth.cust);
      if (!mi)
        return error(t, "custom/undefined method %s used on int or string",
                     t->u.meth.cust);
      return user_method(ei, es, mi, t);
    }
    vn = &t->u.meth.obj->u.var;
    v = lookup_var(vn);
    if (!v)
      return error(t, "custom method used on undefined variable %s",
                    vn->varname);
    return custom_method(ei, es, v, t->u.meth.cust, t);
  }
  if (mte->takes == E_VAR)
  {
    if (t->u.meth.obj->type != T_VAR)
      return error(t, "method %s used on non-variable subject", mte->name);
    vn = &t->u.meth.obj->u.var;
    v = lookup_var(vn);
    if (!v)
      return error(t, "variable %s not defined", vn->varname);
    if (mte->ret == E_INT)
    {
      *ei = mte->func((int)v, t->u.meth.parms);
      return E_INT;
    }
    else
    {
      *es = (char *)mte->func((int)v, t->u.meth.parms);
      return E_STRING;
    }
  }
  er = eval_expr(&thisi, &thiss, t->u.meth.obj);
  if (er == E_ERROR)
    return error(t, "bad subject for method %s", mte->name);
  if (er != mte->takes)
    return error(t, "bad subject type for method %s", mte->name);
  if (mte->ret == E_INT)
  {
    *ei = mte->func(er == E_INT ? thisi : (int)thiss, t->u.meth.parms);
    return E_INT;
  }
  else
  {
    *es = (char *)mte->func(er == E_INT ? thisi : (int)thiss, t->u.meth.parms);
    return E_STRING;
  }
}

/* Variables created or used in the main program body are global i.e.
   zero scope level.  Each user-defined method (function) will push
   a new scope when called.  This scope will contain both formal
   parameters (which are initialised with the values of actual parms)
   and any variables created or used (without the global scope qualifier)
   in the body of the function. 
   
   Each var_node may cache a pointer to the corresponding var instance.
   Since vars may be destroyed (on method return and at end of forall loop),
   we need to keep track of which var_nodes are pointing to the variable
   so that the pointer may be uncached.
   */
#define MAX_RECURSION 100
static hash_table *scope[MAX_RECURSION]; /* Global scope variable hash table */
static var * thisvars[MAX_RECURSION];
static int curr_scope;		/* Index into the above for scope nesting */
#define varht (*(scope[curr_scope]))

void init_var_hash(void)
{
  /* Allocate permanent global scope var hash table */
  scope[curr_scope = 0] = init_ht(0, 477, offsetof(var, name));
}

void push_scope(void)
{
  scope[++curr_scope] = init_ht(0, 19, offsetof(var, name));
}

void pop_scope(void)
{
  if (!curr_scope)
    return;
  /* Iterate through all hash table entries, destroying variables found,
     then delete the hash table itself */
  delete_ht(scope+curr_scope, (void (*)(char *))destroy_var);
  curr_scope--;
}

void pseudo_push_scope(void)
{ curr_scope++; }

void pseudo_pop_scope(void)
{ curr_scope--; }

void set_thisvar(var * v)
{ thisvars[curr_scope] = v; }

static var * add_var_ref(var * v, var_node * vn)
{
  if (!v)
    return 0;
  vn->var = v;
  vn->next = v->refs;
  v->refs = vn;
  return v;
}

var * delete_var_ref(var_node * vn)
{
  /* When recycling parse element, unlink its refs to var. */
  var * v = vn->var;
  var_node * p;
  if (!v)
    return 0;
  if (v->refs == vn)
    v->refs = vn->next;
  else
    for (p = v->refs; p; p = p->next)
      if (p->next == vn)
      {
        p->next == vn->next;
        break;
      }
  return v;
}

var * existing_var(var_node * vn)
{
  /* Return existing var (don't create) */
  if (vn->var)
    return vn->var;
  if (vn->global)
    return add_var_ref((var *)lookup_hash(scope[0], vn->varname), vn);
  return add_var_ref((var *)lookup_hash(&varht, vn->varname), vn);
}

var * lookup_array_var(var_node * vn)
{
  /* Does normal lookup_var.  If this is a scalar, also looks in record
     hash table to see if array was previously created (but went out
     of scope because the first definition was in a method).  If so,
     then the 'scalar' is linked to the array. */
  record * r;
  var * v = lookup_var(vn);
  
  if (!(v->flags & VAR_ARRAY) &&
      (r = (record *)lookup_hash(recht, v->name)) && !r->rec1)
  {
    /* Note that this only needs to be done the first time the var is
       ref'd at the global level, whin initially created in a method. */
    v->flags |= VAR_ARRAY;
    v->rec = r;
    /* Reconstruct index/value types */
    if (r->fields[0].type == FT_STRING)
      v->flags |= VAR_IX_IS_STR;
    if (r->fields[1].type == FT_STRING)
      v->type = E_STRING;
  }
  return v;
}

var * lookup_var(var_node * vn)
{
  /* Locate variable, creating new default if not found.  If name is null,
     is referring to 'this' in a method.  Resolves alias to real var. */
  var * v;
  
  if (vn->var)
  {
    if (vn->var->flags & VAR_ALIAS)
      return vn->var->rvar;
    return vn->var;
  }
  
  if (!vn->varname)
  {
    v = thisvars[curr_scope];	/* NB: if not in a method, v will be null,
    				hence a new plain variable called 'this' will
    				be created. */
    if (!v)
    {
      vn->varname = "this";
      vn->global = 1;
      return lookup_var(vn);
    }
  }
  else
  {
    if (vn->global)
      v = add_var_ref((var *)lookup_hash(scope[0], vn->varname), vn);
    else
      v = add_var_ref((var *)lookup_hash(&varht, vn->varname), vn);
  }
  if (v)
  {
    if (v->flags & VAR_ALIAS)
      return v->rvar;
    return v;
  }
  v = (var *)calloc(sizeof(*v), 1);
  v->name = vn->varname;
  if (vn->global)
    ADD_HASH(scope[0], v);
  else
    ADD_HASH(&varht, v);
  return add_var_ref(v, vn);
}

var * alias_var(var * v, var_node * vn)
{
  /* Create an alias to a variable.  vn->varname is the new name (i.e.
     formal parameter) which will be set to refer to existing variable
     v.  vn->varname must not exist in current scope, otherwise will 
     return null.  Alias variables are just like any other (with
     one or more references from var nodes) except that lookup_var
     dereferences the alias pointer field (rvar). */
  var * w;

  /* Free old alias pointer if any */
  if (vn->alias && vn->var)
    free(vn->var), vn->var = 0, vn->alias = 0;

  w = create_var(vn);
  if (!w)
    return 0;
  vn->alias = 1;
  w->flags |= VAR_ALIAS;
  w->rvar = v;
  return w;
}

var * create_var(var_node * vn)
{
  /* Creating new default variable. If already exists, return 0. */
  var * v = existing_var(vn);
  if (v)
    return 0;
  v = (var *)calloc(sizeof(*v), 1);
  v->name = vn->varname;
  if (vn->global)
    ADD_HASH(scope[0], v);
  else
    ADD_HASH(&varht, v);
  /*fprintf(stderr, "Created var `%s' at %08X\n", name, v);*/
  return add_var_ref(v, vn);
}

void init_defs(int ndefs, char ** defs)
{
  /* Initialise variables based on -D or -N command line options.  Def is in
     form [D|N]varname=value or just varname.  In the latter case, the value
     is set to integer 1.  For assigned values, a string or integer value is 
     set: string for D, integer for N. */
  int i;
  var * v;
  var_node vn;
  char * val;
  
  vn.global = 1;
  for (i = 0; i < ndefs; i++)
  {
    /*fprintf(stderr,"Creating def %d = `%s'...\n", i, defs[i]);*/
    if ((val = strchr(defs[i], '=')))
      *val++ = 0;
    vn.varname = defs[i]+1;
    vn.var = 0;
    v = create_var(&vn);
    if (!v)
    {
      if (verbose)
        fprintf(stderr, "Redefinition of `%s', using original.\n", defs[i]);
      continue;
    }
    /*fprintf(stderr,"`%s' ok.\n", vn.varname);*/
    v->refs = 0;
    if (val)
    {
      char tchar = defs[i][0];
      if (tchar == 'D')
        set_var(v, E_STRING, 0, val);
      else
      {
        int x;
        if (val[0] == '0' && toupper(val[1]) == 'X')
          sscanf(val+2, "%x", &x);
        else
          sscanf(val, "%d", &x);
        set_var(v, E_INT, x, 0);
      }
    }
    else
      set_var(v, E_INT, 1, 0);
  }
}

void set_var(var * v, expr_result er, int ei, char * es)
{
  if (v->flags & VAR_READONLY)
    return;
  v->type = er;
  v->intval = ei;
  if (er == E_STRING)
  {
    int l = strlen(es) + 1;
    if (l > v->alloc)
    {
      if (v->strval)
        free(v->strval);
      v->strval = (char *)malloc(v->alloc = l + 16);
    }
    strcpy(v->strval, es);
  }
}

void set_array(var * v, tree t, expr_result erx, int eix, char * esx, 
                        expr_result er, int ei, char * es)
{
  object * o;
  record * r;
  
  if (v->flags & VAR_SLICE)
  {
    int ool; /* Set if slice is out of line of original allocation */
    int ml;  /* Max length of new val if to remain in original alloc 
                (0 if ool) */
    char * p;
    int l;
    
    if (erx != E_INT)
    {
      error(t, "index of slice var %s must be integer", v->name);
      return;
    }
    if (er != E_STRING)
    {
      error(t, "value of slice var %s must be string", v->name);
      return;
    }
    if (eix < 0 || eix > v->intval * 2)
    {
      /* We don't currently expand the number of slices */
      error(t, "slice var %s has no part number %d", v->name, eix);
      return;
    }
    p = v->si[eix].begin;
    if (p >= v->strval && p < v->strval + v->alloc)
      ml = v->si[eix].len;  /* We cannot expand yet */
    else
      p = 0, ml = 0;
    l = strlen(es);
    if (p && l <= ml)
      memcpy(p, es, l);
    else
    {
      if (!p && v->si[eix].begin)
        free(v->si[eix].begin);
      p = (char *)malloc(l);
      memcpy(p, es, l);
      v->si[eix].begin = p;
    }
    v->si[eix].len = l;
    return;
  }
  if (!(v->flags & VAR_ARRAY))
  {
    v->flags |= VAR_ARRAY;  /* Convert erstwhile scalar to array.  Original
                               scalar type and value are discarded. */
    if ((r = (record *)lookup_hash(recht, v->name)) && r->rec1)
    {
      error(t, "array name %s conflicts with database", v->name);
      return;
    }
    if (!r)
    {
      v->rec = create_record_type(recht, v->name, 2,
              CRT_DEFINITIVE | (erx == E_STRING ? 0 : CRT_INT), "index", 
              er == E_STRING ? 0 : CRT_INT, "value");
      if (!v->rec)
      {
        error(t, "could not create array %s", v->name);
        return;
      }
      if (erx == E_STRING)
        v->flags |= VAR_IX_IS_STR;
      v->type = er;
    }
    else
    {
      v->rec = r;  /* This happend if array was created inside a method
                      thus its var went out of scope.  We reset the
                      record pointer here */
      /* Reconstruct index/value types */
      if (r->fields[0].type == FT_STRING)
        v->flags |= VAR_IX_IS_STR;
      if (r->fields[1].type == FT_STRING)
        v->type = E_STRING;
    }
  }
  if (er != v->type)
  {
    error(t, "array value type mismatch on assignment to %s", v->name);
    return;
  }
  if (erx == E_STRING ^ !!(v->flags & VAR_IX_IS_STR))
  {
    error(t, "array index type mismatch on assignment to %s", v->name);
    return;
  }
  if (erx == E_INT)
    esx = (char *)&eix;
  if (er == E_STRING)
  {
      /* Allocate private copy of value string */
      int l = strlen(es);
      char * s = (char *)malloc(l+1);
      memcpy(s, es, l);
      s[l] = 0;
      es = s;
  }
  o = (object *)lookup_hash(&v->rec->objht, esx);
  if (!o)
  {
    if (erx == E_STRING)
    {
      /* Allocate private copy of index string */
      int l = strlen(esx);
      char * s = (char *)malloc(l+1);
      memcpy(s, esx, l);
      s[l] = 0;
      esx = s;
    }
    o = create_object(v->rec, 2, 
             erx == E_STRING ? esx : (char *)eix, 
             er == E_STRING ? es : (char *)ei);
  }
  else
  {
    if (v->type == E_STRING && o->data[1])
      free(o->data[1]);  /* Free old private copy of value */
    o->data[1] = er == E_STRING ? es : (char *)ei;
  }
}

static char * m_fieldval(var * v, list parms)
{
  meth_tab_ent * mte = v->mte;
  field * f = (field *)mte->desc;
  object * o = v->obj;
  
  /*fprintf(stderr, "fieldval `%s' of var `%s' at %08X\n", f->name, v->name, v);
  */
  if (f->type == FT_INT || f->type == FT_STRING || 
      !(o->resolved & 1 << f->fldno))
    return o->data[f->fldno];
  o = (object *)o->data[f->fldno];
  return o->data[f->ref->definitive];
}

expr_result array_lookup(int * ei, char ** es, var * v, tree t,
                         expr_result erx, int eix, char * esx)
{
  record * r;
  object * o;
  
  if (v->flags & VAR_SLICE)
  {
    if (erx != E_INT)
    {
      error(t, "index of slice var %s must be integer", v->name);
      return E_ERROR;
    }
    if (eix < 0 || eix > v->intval * 2)
      *es = "";
    else
    {
      *es = tsmalloc(v->si[eix].len + 1);
      memcpy(*es, v->si[eix].begin, v->si[eix].len);
      (*es)[v->si[eix].len] = 0;
    }
    return E_STRING;
  }
  r = v->rec;
  if (!r)
  {
    r = (record *)lookup_hash(recht, v->name);
    if (!r)
      return error(t, "array %s not found", v->name);
    v->rec = r;
  }
  if (erx == E_INT)
  {
    if (v->flags & VAR_IX_IS_STR)
      return error(t, "array %s index must be string", v->name);
    esx = (char *)&eix;
  }
  else
  {
    if (!(v->flags & VAR_IX_IS_STR))
      return error(t, "array %s index must be integer", v->name);
  }
  o = (object *)lookup_hash(&r->objht, esx);
  if (!o)
  {
    *ei = 0;
    *es = "";
  }
  else
  {
    if (v->type == E_INT)
      *ei = (int)o->data[1];
    else
      *es = o->data[1];
  }
  return v->type;
}

#define CHT_SLOTS 19
void init_forallvar(var * v, record * r)
{
  field * f;
  unsigned nf;
  meth_tab_ent * mte;
  
  v->flags |= VAR_READONLY | VAR_FORALL;
  v->rec = r;
  v->cht = init_ht(0, CHT_SLOTS, offsetof(meth_tab_ent, name));
  v->mtes = mte = (meth_tab_ent *)calloc(sizeof(meth_tab_ent) * r->nfields, 1);
  for (nf = 0; nf < r->nfields; nf++)
  {
    f = r->fields + nf;
    mte[nf].ret = f->type == FT_INT ? E_INT : E_STRING;
    mte[nf].func = (int (*)(int, list))&m_fieldval;
    mte[nf].name = f->name;
    mte[nf].desc = (char *)f;
    ADD_HASH(v->cht, mte + nf);
  }
  if (r->definitive >= 0)
    v->type = E_STRING;
  else
    v->type = E_INT;
  v->intval = -1;
  /*fprintf(stderr, "init_forallvar var `%s' at %08X, rec `%s'\n", 
    v->name, v, r->name);*/
}

field * init_nested_forallvar(var * v, var * cv, char * fldname)
{
  field * f = 0;
  referback * rb;
  
  /*fprintf(stderr, "init_nested_forallvar: fldname = `%s' of record `%s'\n",
    fldname, cv->rec->name);*/
  
  for (rb = cv->rec->rbacks; rb; rb = rb->next)
    if (!strcmp(rb->fld->owner->name, fldname))
      break;
  if (rb)
  {
    f = rb->fld;
    init_forallvar(v, f->owner);
  }
  /*fprintf(stderr, "init_nested_forallvar var `%s' at %08X, rec `%s'\n"
                  "   cv = `%s' at %08X\n",
    v->name, v, cv->rec->name, cv->name, cv); */
  return f;
}

void set_forallvar(var * v, object * o)
{
  v->obj = o;
  v->intval++;
  if (v->rec->definitive >= 0)
    v->strval = o->data[v->rec->definitive];
  /*fprintf(stderr, "set_forallvar var `%s' at %08X\n", v->name, v);*/
}

void destroy_var(var * v)
{
  /*fprintf(stderr, "destroy_var var `%s' at %08X\n", v->name, v);*/
  /* First, reset all cached entries pointing to this var. */
  var_node * vn, * vn2;
  if (v->flags & VAR_SLICE)
  {
    char * p;
    int i;
    /* Free out-of-line slices */
    for (i = 0; i < v->nslice; i++)
    {
      p = v->si[i].begin;
      if (p < v->strval || p >= v->strval + v->alloc)
        free(p);
    }
  }
  for (vn = v->refs; vn; vn = vn2)
  {
    vn2 = vn->next;
    vn->var = 0;
    vn->next = 0;
  }
  delete_ht(&v->cht, 0);
  if (v->alloc && v->strval)
    free(v->strval);
  if (v->mtes)
    free(v->mtes);
  if (v->flags & VAR_SLICE && v->si)
    free(v->si);
  free(v);
}

void unhash_destroy_var(var * v)
{
  /*fprintf(stderr, "destroy_var var `%s' at %08X\n", v->name, v);*/
  unhash(&varht, (char *)v);
  destroy_var(v);
}

#define TS_INIT 65536	/* Initial and minimum increment temp stack area */
static char * ts_base;	/* Base of temp stack area */
static long ts_alloc;	/* Current allocated size */
static long ts_top;	/* Current next free */
static long ts_hwm;	/* High water mark */
char ** ts_topmark;	/* Pointer to topmost mark in temp stack */

char * tsmalloc(long size)
{
  char * ret;
  /* Round up size to mult of 4 to avoid bus error on word-addressed mach */
  size = (size-1 | 3) + 1;
  if (ts_top + size > ts_alloc)
  {
    if (!ts_base)
      ts_base = (char *)malloc(ts_alloc = TS_INIT);
    else
    {
      fprintf(stderr, "Temp stack overflow at %d bytes\n", ts_alloc);
      exit(5);
    }
  }
  ret = ts_base + ts_top;
  ts_top += size;
  if (ts_top > ts_hwm)
    ts_hwm = ts_top;
  return ret;
}

long mark_tempstack(void)
{
  return ts_top;
}

void pop_tempstack(long mark)
{
  ts_top = mark;
}

long get_hwm(void)
{
  return ts_hwm;
}
  

char * make_permanent(char ** s, int * alloc)
{
  
}

