/*****************************************************************************
*   $Id: dgvar.h,v 1.1.1.1 2000/06/22 15:39:56 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    techno@spirit.net.au
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: dgvar.h,v $
* Revision 1.1.1.1  2000/06/22 15:39:56  tekno
* Opened source
*
*
*****************************************************************************/
/* dgvar.h - variable struct def. */

typedef struct meth_tab_ent_t
{
  expr_result ret;
  expr_result takes;
  int (*func)(int, list);
  int nparms;
  char * name;
  char * desc;
} meth_tab_ent;

typedef struct slice_info_t
{
  char * begin;
  unsigned long len;
} slice_info;

typedef struct var_t
{
  char * name;
  unsigned long flags;	/* Flag bits as follows: */
#define VAR_READONLY	0x01 /* If set, this is readonly var (to program). */
#define VAR_FORALL	0x02 /* If set, this is a FORALL loop control var */
#define VAR_FIRST	0x04 /* This is first iteration */
#define VAR_LAST	0x08 /* This is last iteration */
#define VAR_SLICE	0x10 /* This var is result of regexp slice/parts meth.*/
#define VAR_ARRAY	0x20 /* This is array variable */
#define VAR_IX_IS_STR	0x40 /* Array variable index is string type */
#define VAR_ALIAS       0x80 /* This is an alias, real var is pointed to
                                by rvar */

  expr_result type;	/* E_INT or E_STRING: current type (defaults to INT).
                           For arrays, this is the type of the values therein.*/
  int intval;		/* Value if integer */
  char * strval;	/* Value if string.  This buffer is malloc'd if
  			   alloc is non-zero, otherwise is shared. */
  int alloc;		/* Allocated size of the above.  Expands if
  			   necessary, but doesn't shrink */
  struct var_node_t * refs;	/* Point to list of var_nodes which have
  				   cached the address of this var. */
  /* The following fields are only valid for 'forall' loop control vars or
     slice vars. */
  meth_tab_ent * mtes;	/* Current method table */
  meth_tab_ent * mte;	/* Current method table entry for custom */
  struct object_t * obj;/* If forall var, point to current object */
  struct hash_table_t * cht;	/* Custom method hash table */
  struct record_t * rec;	/* Object type iterating for, or points
                                   to constructed database if this is an
                                   array variable. */
  /* Foll. fields only valid for slice vars */
  int nslice;			/* Number of entries in foll. */
  slice_info * si;		/* Substrings.  These entries point into
  				   the strval buffer. */
  struct var_t * rvar;		/* Original variable if this is an alias */
} var;

typedef struct method_info_t
{
  /* User-defined methods */
  char * name;		/* Method name */
  tree func;		/* T_FUNCTION node used to define this method */
} method_info;


struct record_t;
struct object_t;
struct var_node_t;

/* Variable access in dgvar.c */
void init_var_hash(void);
void push_scope(void);
void pop_scope(void);
void pseudo_push_scope(void);
void pseudo_pop_scope(void);
void set_thisvar(var * v);
var * delete_var_ref(var_node * vn);
var * existing_var(struct var_node_t * vn);
var * lookup_array_var(struct var_node_t * vn);
var * lookup_var(struct var_node_t * vn);
var * alias_var(var * v, var_node * vn);
var * create_var(struct var_node_t * vn);
void set_var(var * v, expr_result er, int ei, char * es);
void set_array(var * v, tree t, expr_result erx, int eix, char * esx, 
                        expr_result er, int ei, char * es);
expr_result array_lookup(int * ei, char ** es, var * v, tree t,
                         expr_result erx, int eix, char * esx);
void init_forallvar(var * v, struct record_t * r);
struct field_t * init_nested_forallvar(var * v, var * cv, char * fldname);
void set_forallvar(var * v, struct object_t * o);
void destroy_var(var * v);
void unhash_destroy_var(var * v);

/* User-defined methods in dgmeth.c */
method_info * get_usermethod(char * mn);
int grok_mdef(tree md);
int grok_mdefs(list root);

/* Storage management */
char * tsmalloc(long size);
long mark_tempstack(void);
void pop_tempstack(long mark);
long get_hwm(void);
