/*****************************************************************************
*   $Id: main.c,v 1.3 2002/02/17 21:50:23 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    techno@spirit.net.au
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: main.c,v $
* Revision 1.3  2002/02/17 21:50:23  tekno
* Rebuild
*
* Revision 1.2  2000/06/29 13:13:13  tekno
* Added -c option to allow commenting out of -m macros
*
* Revision 1.1.1.1  2000/06/22 15:39:59  tekno
* Opened source
*
*
*****************************************************************************/
/* main.c - mainline for COOK project */

#include <stdio.h>
#include <stdlib.h>
#include "dglex.h"
#include "dgvar.h"
#include "dgdb.h"

int pass1error;
int xdebug = 1;
int free_string = 0;

#define MAX_DATABASES   100
#define MAX_DEFINES     100

int main(int argc, char ** argv)
{
  int ret;
  extern FILE * yyin, * yyout;
  extern int yyparse(void);
  extern int execute(list);
  int i;
  char tchar;	/* 'D' or 'N' depending on whether -D or -N arg */
  
  char * incdir;
  char * cmntstr;
  
  char * dbases[MAX_DATABASES];
  char * dbase;
  int ndbases = 0;
  
  char * defs[MAX_DEFINES];
  char * def;
  int ndefs = 0;
  
  
  
  char * progname = argv[0];
  
  srcfilename = 0;
  
  /*for (i = 0; i < argc; i++)
    fprintf(stderr, "argv[%d] = `%s'\n", i, argv[i]);*/
  
  cmntstrs[ncmntstrs++] = "//";	/* Set default line comment strings */
  cmntstrs[ncmntstrs++] = ";";
  
  while (argc > 1)
  {
    if (!strcmp(argv[1], "-v"))
      verbose = 1;
    else if (!strcmp(argv[1], "+v"))
      verbose = 0;
    else if (!strcmp(argv[1], "-r"))
      raw = 1;
    else if (!strcmp(argv[1], "+r"))
      raw = 0;
    else if (!strcmp(argv[1], "-f"))
      free_string = 1;
    else if (!strcmp(argv[1], "+f"))
      free_string = 0;
    else if (!strcmp(argv[1], "-p"))
      preparse = 1;
    else if (!strcmp(argv[1], "+p"))
      preparse = 0;
    else if (!strcmp(argv[1], "-l"))
      lbl = 1;
    else if (!strcmp(argv[1], "+l"))
      lbl = 0;
    else if (!strcmp(argv[1], "-#"))
      srcident = 1;
    else if (!strcmp(argv[1], "+#"))
      srcident = 0;
    else if (!strcmp(argv[1], "+c"))
      ncmntstrs = 0;	/* Discard all line comment strings */
    else if (!strncmp(argv[1], "-i", 2))
    {
      if (strlen(argv[1]) > 2)
        sscanf(argv[1]+2, "%d", &maxsrcfiles);
      else
        sscanf(argv[2], "%d", &maxsrcfiles), argv++, argc--;
    }
    else if (!strncmp(argv[1], "-j", 2))
    {
      if (strlen(argv[1]) > 2)
        sscanf(argv[1]+2, "%d", &maxsrcnest);
      else
        sscanf(argv[2], "%d", &maxsrcnest), argv++, argc--;
    }
    else if (!strncmp(argv[1], "-+", 2))
    {
      /* As of 1.0.1, this option changed to undocumented.  Used to be -c
         for line continuation, but this is unimplemented.  Now use -c for
         line comment.  This option changed to -+ */
      if (strlen(argv[1]) > 2)
        cont_str = argv[1]+2;
      else
        cont_str = argv[2], argv++, argc--;
    }
    else if (!strncmp(argv[1], "-m", 2))
    {
      if (strlen(argv[1]) > 2)
        macro_str = argv[1]+2;
      else
        macro_str = argv[2], argv++, argc--;
    }
    else if (!strncmp(argv[1], "-I", 2))
    {
      if (nincdirs == MAX_INCDIRS)
      {
        fprintf(stderr, "More than %d include directories specified\n",
          MAX_INCDIRS);
        return 2;
      }
      if (strlen(argv[1]) > 2)
        incdir = argv[1]+2;
      else
        incdir = argv[2], argv++, argc--;
      incdirs[nincdirs] = (char *)malloc(strlen(incdir)+1);
      strcpy(incdirs[nincdirs++], incdir);
      /*fprintf(stderr, "Incdir: `%s'\n", incdir);*/
    }
    else if (!strncmp(argv[1], "-c", 2))
    {
      if (ncmntstrs == MAX_COMMENT_STRS)
      {
        fprintf(stderr, "More than %d line comment strings specified\n",
          MAX_COMMENT_STRS);
        return 2;
      }
      if (strlen(argv[1]) > 2)
        cmntstr = argv[1]+2;
      else
        cmntstr = argv[2], argv++, argc--;
      cmntstrs[ncmntstrs] = (char *)malloc(strlen(cmntstr)+1);
      strcpy(cmntstrs[ncmntstrs++], cmntstr);
    }
    else if (!strncmp(argv[1], "-d", 2))
    {
      if (ndbases == MAX_DATABASES)
      {
        fprintf(stderr, "More than %d database files specified\n",
          MAX_DATABASES);
        return 2;
      }
      if (strlen(argv[1]) > 2)
        dbase = argv[1]+2;
      else
        dbase = argv[2], argv++, argc--;
      dbases[ndbases] = (char *)malloc(strlen(dbase)+1);
      strcpy(dbases[ndbases++], dbase);
      /*fprintf(stderr, "Database: `%s'\n", dbase);*/
    }
    else if (!strncmp(argv[1], "-D", 2) || !strncmp(argv[1], "-N", 2))
    {
      tchar = argv[1][1];
      if (strlen(argv[1]) > 2)
        def = argv[1]+2;
      else
        def = argv[2], argv++, argc--;
    add_def:
      if (ndefs == MAX_DEFINES)
      {
        fprintf(stderr, "More than %d defines specified\n", MAX_DEFINES);
        return 2;
      }
      if (!isalpha(def[0]))
      {
        fprintf(stderr, "Definition var name invalid\n");
        return 2;
      }
      defs[ndefs] = (char *)malloc(strlen(def)+2);
      defs[ndefs][0] = tchar;
      strcpy(defs[ndefs++]+1, def);
    }
    else if ((*argv[1] != '-' || !strcmp(argv[1], "-")) && !srcfilename)
    {
      srcfilename = argv[1];
    }
    else if (isalpha(*argv[1]) && srcfilename)
    {
      def = argv[1];
      tchar = 'D';
      goto add_def;
    }
    else
    {
      fprintf(stderr,
 "COOK Macro Processor V1.0.2  (C) 2000 SJH\n"
 "usage: %s [-v] [-r] [-f] [-l [-c str]... [-m char]] [-#] [-p]\n"
 "   [-d db_file]... [-Dvar[=val]]... [-Idir]... [-i n] [-j n] [src_file]\n"
 "	-v      Verbose\n"
 "	-r      Raw output i.e. ignore filters\n"
 "	-f      Start in free-string mode\n"
 "	-l      Line-by-line strings...\n"
 "	  -c      Line comment string (default ; and //)  +c to turn off\n"
 "	  -m      Macro recognition char (default no macros)\n"
 "	-#      Emit source file identification (#line directives)\n"
 "	-p      Parse all before execution (else execute ASAP)\n"
 "	-d      Specify database file(s)\n"
 "	-D      Specify initialised string variables.  Also:\n"
 "	         -Nvar[=val]  Set variable to integer value.\n"
 "	-I      Specify directories for #include <file>\n"
 "	-i      Specify maximum number of distinct source files (dflt 255)\n"
 "	-j      Specify maximum source file nesting (dflt 10)\n"
 "	src_file  Program to execute (- for stdin)\n\n"
 "  Any args after src_file are treated as parameters of -D\n"
             , progname);
      return 1;
    }
    
    argv++;
    argc--;
  }
  
  nrec = 0;
  ret = 0;
  
  if (maxsrcfiles < 1)
    maxsrcfiles = 1;
  if (maxsrcfiles > 65536)
    maxsrcfiles = 65536;
  if (maxsrcnest < 1)
    maxsrcnest = 1;
  if (maxsrcnest > 256)
    maxsrcnest = 256;
  
  for (i = 0; i < ndbases; i++)
  {
    /*fprintf(stderr, "Database %d: `%s'\n", i, dbases[i]);*/
    if (i == ndbases - 1)
    {
      if ((ret = read_db(dbases[i], &nrec, &recht, &rpt)))
        return ret;
    }
    else if ((ret = read_db(dbases[i], 0, 0, 0)))
      return ret;
  }
  if (!recht)
    /* There were no databases.  Set up an empty database for future array
       definitions etc. */
    init_db(&recht);
  
  if (srcfilename && strcmp(srcfilename, "-"))
  {
    yyin = fopen(srcfilename, "r");
    if (!yyin)
    {
      fprintf(stderr, "Could not open source file `%s' for reading\n", 
        srcfilename);
      return 2;
    }
  }
  else
  {
    yyin = stdin;
    srcfilename = "-";
  }
  outfile = stdout;	/* Default starting output file */
  init_var_hash();
  init_defs(ndefs, defs);
  if (init_lex(free_string, 1))
  {
    fprintf(stderr, "Could not initialise lexical scanner\n");
    return 20;
  }
  if (!preparse)
    grok_mdefs(0);	/* Init method hash table, but nothing else */
  yyparse();
  fclose(yyin);
  if(pass1error)
  {
    if (preparse)
      fprintf(stderr, "Pass 1 (syntax analysis) failed\n");
    return 3;
  }
  if (preparse)
  {
    ret = grok_mdefs(tmpl_root); /* Methods may be forward */
    if (!ret)
    {
      ret = execute(tmpl_root);
      flush_filters();
    }
  }
  if (verbose)
    fprintf(stderr, "Executed: rc = %d, stack used = %ld\n", ret, get_hwm());
  return ret;
}
