/*****************************************************************************
*   $Id: ts.c,v 1.1.1.1 2000/06/22 15:40:02 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    techno@spirit.net.au
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: ts.c,v $
* Revision 1.1.1.1  2000/06/22 15:40:02  tekno
* Opened source
*
*
*****************************************************************************/
/* testscan.c - test lexical scanner */

#include <stdio.h>
#include "dglex.h"
#include "y.tab.h"




int main(int argc, char ** argv)
{
  int tk;
  char fn[257];
  
  if (argc < 2)
  {
    yyin = stdin;
    strcpy(fn, "<stdin>");
  }
  else
  {
    yyin = fopen(argv[1], "r");
    if (!yyin)
    {
      perror(argv[1]);
      return 2;
    }
    strcpy(fn, argv[1]);
  }
  srcfilename = fn;
  lbl = 1;
  macro_str = "$";
  init_lex(1, 1);
  do
  {
    tk = yylex();
    printf("Token %d at line %d  length %d:\n", tk, lex_tkstart, yyleng);
    switch (tk)
    {
    case INTEGER:
      printf("  int %d\n", yylval.lex_int);
      break;
    case FLOATING:
      printf("  float %lg\n", yylval.lex_float);
      break;
    case STRING:
      printf("  string\n");
      fputs(yylval.lex_str, stdout);
      free(yylval.lex_str);
      printf("<---(end of string)---\n");
      break;
    case VAR:
      printf("  var %s\n", yylval.lex_str);
      break;
    case IF:
      puts("if");
      break;
    case ELSE:
      puts("else");
      break;
    case ELSEIF:
      puts("elseif");
      break;
    case ENDIF:
      puts("endif");
      break;
    case FOR:
      puts("for");
      break;
    case FORALL:
      puts("forall");
      break;
    case NEXT:
      puts("next");
      break;
    case KEEP:
      puts("keep");
      break;
    case IFNONE:
      puts("ifnone");
      break;
    case SORT:
      puts("sort");
      break;
    case SWITCH:
      puts("switch");
      break;
    case CASE:
      puts("case");
      break;
    case ENDSWITCH:
      puts("endswitch");
      break;
    case METHOD:
      puts("method");
      break;
    case RETURN:
      puts("return");
      break;
    case THIS:
      puts("this");
      break;
    case GLOBAL:
      puts("::");
      break;
    case OUTFILE:
      puts("outfile");
      break;
    case GE:
      puts(">=");
      break;
    case LE:
      puts("<=");
      break;
    case NE:
      puts("!=");
      break;
    case LOGAND:
      puts("&&");
      break;
    case LOGOR:
      puts("||");
      break;
    case MATCHES:
      puts("~");
      break;
    case STRICTMATCHES:
      puts("~=");
      break;
    case EQ:
      puts("==");
      break;
    case DBLSEMI:
      puts(";;");
      break;
    case REFERS:
      puts("<-");
      break;
    default:
      if (isprint(tk))
        printf("%c\n", tk);
      else
        printf("  unknown %d\n", tk);
      break;
    }
  } while (tk);
  
  return 0;
}
